from .types import _SubparserType


def add_sandbox_command(subparsers: _SubparserType):
    sandbox_test_parser = subparsers.add_parser("sandbox", help="Test a mindgard example model")
    sandbox_test_parser.add_argument("--json", action="store_true", help="Return json output", required=False)
    sandbox_test_parser.add_argument(
        "--risk-threshold",
        type=int,
        help="Set a flagged event to total event ratio threshold above which the system will exit 1",
        required=False,
        default=80,
    )
    return sandbox_test_parser



# Standard library imports
import textwrap

from ..types import valid_llm_datasets
from .shared import add_shared_arguments
from .types import _SubparserType


def add_test_command(subparsers: _SubparserType):
    test_parser = subparsers.add_parser("test", help="Attacks command")
    add_shared_arguments(test_parser)
    test_parser.add_argument(
        "--parallelism",
        type=int,
        help="The maximum number of parallel requests that can be made to the API.",
        required=False,
    )
    test_parser.add_argument(
        "--rate-limit",
        type=int,
        help="The maximum number of requests to make to model in one minute (default: 3600)",
        required=False,
    )
    test_parser.add_argument(
        "--force-multi-turn",
        type=bool,
        help="Enable multi turn attacks in scenarios where they may not be safe, such as when testing an API without chat completions history.",
        required=False,
    )
    test_parser.add_argument(
        "--dataset",
        type=str,
        help=textwrap.dedent(
            f"""
            The dataset to be used for running the attacks on the given model.
            This should be a csv formatted file path, with each prompt on a new line"""
        ),
        required=False,
    )
    test_parser.add_argument(
        "--domain",
        type=str,
        help="The domain to inform the dataset used for LLMs.",
        choices=valid_llm_datasets,
        required=False,
    )
    test_parser.add_argument(
        "--mode",
        type=str,
        help="Specify the number of samples to use during attacks; contact Mindgard for access to 'thorough' or 'exhaustive' test",
        choices=["fast", "thorough", "exhaustive"],
        required=False,
    )
    test_parser.add_argument(
        "--exclude",
        type=str,
        help=textwrap.dedent(
            f"""
            Exclude certain attacks from the test. Exclusions can be done either by name or category.
            The supported attacks can be found here - https://docs.mindgard.ai/user-guide/running-subset-of-attacks#list-of-attacks"""
        ),
        action="append",
        required=False,
    )
    test_parser.add_argument(
        "--include",
        type=str,
        help=textwrap.dedent(
            f"""
            Include a selected set of attacks in the test. A name or category can be provided as part of the inclusion. 
            The supported attacks can be found here - https://docs.mindgard.ai/user-guide/running-subset-of-attacks#list-of-attacks"""
        ),
        action="append",
        required=False,
    )
    test_parser.add_argument(
        "--prompt-repeats",
        type=int,
        help="The number of times to repeat the prompt for each sample in the dataset.",
        required=False,
    )
    return test_parser



# Standard library imports
import logging
from typing import List

# Third party imports
import requests
from pydantic import BaseModel

# Project imports
from mindgard.api.exceptions import ClientException
from mindgard.version import VERSION


class CreateProjectRequest(BaseModel):
    """
    Request model for creating a project.

    Attributes:
        project_name (str): The name of the project to create.
    """

    project_name: str


class CreateProjectResponse(BaseModel):
    """
    Response model for creating a project.

    Attributes:
        id (str): Alphanumeric identifier for the created project.
    """

    id: str
    name: str


class ProjectItem(BaseModel):
    """
    Model representing a single project item.
    """

    id: str
    name: str


class ListProjectsResponse(BaseModel):
    """
    Response model for listing projects.

    Attributes:
        items (List[ProjectItem]): List of project items.
    """

    items: List[ProjectItem]


class ProjectClient:
    """
    Client for managing multiturn attack sessions via the API.

    Inherits from AttackPromptEventsClient and provides methods to start and interact
    with multiturn attack sessions, handling authentication and request formatting.

    Args:
        base_url (str): The base URL of the API endpoint.
        access_token (str): Bearer token for authentication.
    """

    def __init__(self, base_url: str, access_token: str):
        """
        Initialize the AttackMultiturnClient.

        Args:
            base_url (str): The base URL of the API endpoint.
            access_token (str): Bearer token for authentication.
        """
        self._base_url = base_url
        self._projects_url = f"{base_url.rstrip('/')}/projects"
        self._access_token = access_token
        self._session = requests.Session()
        self._session.headers.update(
            {
                "Authorization": f"Bearer {self._access_token}",
                "Content-Type": "application/json",
                "User-Agent": f"mindgard-cli/{VERSION}",
                "X-User-Agent": f"mindgard-cli/{VERSION}",
            }
        )

    def create(self, req: CreateProjectRequest) -> CreateProjectResponse:
        """
        Creates a new project.

        Args:
            req (CreateProjectRequest): The request containing project name.

        Returns:
            CreateProjectResponse: The response containing the ID of the created project.

        Raises:
            ClientException: If the request fails or response validation fails.
        """
        response = self._session.post(self._projects_url, data=req.model_dump_json())

        if response.status_code != 201:
            logging.debug(f"Failed to create project: {response.json()} - {response.status_code}")
            raise ClientException(response.text, response.status_code)

        try:
            return CreateProjectResponse.model_validate(response.json())
        except Exception as e:
            raise ClientException("Problem parsing CreateProjectResponse", response.status_code) from e

    def list(self) -> ListProjectsResponse:
        """
        Lists all projects.

        Returns:
            ListProjectsResponse: The response containing the list of projects.

        Raises:
            ClientException: If the request fails or response validation fails.
        """
        response = self._session.get(self._projects_url)

        if response.status_code != 200:
            logging.debug(f"Failed to get projects: {response.json()} - {response.status_code}")
            raise ClientException(response.text, response.status_code)

        try:
            return ListProjectsResponse.model_validate(response.json())
        except Exception as e:
            raise ClientException("Problem parsing CreateProjectResponse", response.status_code) from e



# Third party imports
from rich.console import Console
from rich.table import Table

# Project imports
from mindgard.api.exceptions import ClientException
from mindgard.api.project.project import CreateProjectRequest, ProjectClient
from mindgard.auth import require_auth
from mindgard.constants import API_BASE, DASHBOARD_URL
from mindgard.utils import CliResponse

console = Console()


class CreateProjectCLIArgs:
    name: str


@require_auth
def create_project(create_project_args: CreateProjectCLIArgs, access_token: str):
    """
    Create a new project with the given name.
    """
    project_service = ProjectClient(API_BASE, access_token)

    try:
        with console.status("Creating project...", spinner="dots"):
            response = project_service.create(CreateProjectRequest(project_name=create_project_args.name))
    except ClientException as ex:
        console.print(f"[red bold]Error: {ex.message}[/red bold]")
        exit(CliResponse(ex.status_code).code())

    console.print(f"\nProject created successfully!\n", style="bold green")
    console.print(f"Project name: {response.name}")
    console.print(f"Project ID: {response.id}\n")
    console.print(
        f"Add the project ID to your config file or pass it with the command (e.g., --project-id {response.id} )."
    )
    console.print("Visit https://docs.mindgard.ai/user-guide/projects for more details.\n")


@require_auth
def list_projects(access_token: str):
    """
    Create a new project with the given name.
    """
    project_service = ProjectClient(API_BASE, access_token)

    try:
        with console.status("Fetching all your projects...", spinner="dots"):
            response = project_service.list()
    except ClientException as ex:
        console.print(f"[red bold]Error: {ex.message}[/red bold]")
        exit(CliResponse(ex.status_code).code())

    if len(response.items) == 0:
        console.print(
            "\nNo projects available. Run `mindgard create project --name <New Project Name>` to create a project"
        )
        return

    table = Table(title="Projects")
    table.add_column("Project ID", style="blue", overflow="fold")
    table.add_column("Project Name", style="white", overflow="fold")
    table.add_column("Project URL", style="white", overflow="fold")

    for _, item in enumerate(response.items, start=1):
        table.add_row(item.id, item.name, f"{DASHBOARD_URL}/results/tests?project_id={item.id}")

    console.print(table)



# Standard library imports
import logging
from typing import Tuple

# Third party imports
from rich.console import Console

from .exceptions import HTTPBaseError, MGException, Uncontactable
from .wrappers.llm import LLMModelWrapper


def preflight_llm(
    model_wrapper: LLMModelWrapper,
    console: Console,
    json_out: bool,
) -> bool:
    """
    Makes requests to the LLM to validate basic connectivity before submitting
    test.

    Returns True on success, False on failure
    """
    try:
        if not json_out:
            console.print("[white]Contacting target...")
        model_wrapper.__call__("Hello llm, are you there?")
        return True
    except Uncontactable as cerr:
        logging.debug(cerr)
        model_api = model_wrapper.api_url if hasattr(model_wrapper, "api_url") else "<unknown>"
        if not json_out:
            console.print(
                f"[red]Could not connect to the target! [white](URL: {model_api}, are you sure it's correct?)"
            )
    except HTTPBaseError as httpbe:
        logging.debug(httpbe)
        message: str = f"[red]Target pre-flight check returned {httpbe.status_code} ({httpbe.status_message})"
        if httpbe.status_code == 401:
            message += " Is your API token correct/valid/permissive?"
        if not json_out:
            console.print(message)
        raise MGException(f"{httpbe}")
    except Exception as e:
        # something we've not really accounted for caught
        logging.error(e)
        raise e

    return False



from .shared import add_shared_arguments
from .types import _SubparserType


def add_validate_command(subparsers: _SubparserType):
    validate_parser = subparsers.add_parser("validate", help="Validates that we can communicate with your model")
    add_shared_arguments(validate_parser)
    return validate_parser



# Standard library imports
from typing import Any, Dict, Optional

# Third party imports
from azure.messaging.webpubsubclient.models import OnGroupDataMessageArgs
from pydantic import BaseModel, model_validator

from .types import type_wps_message_type


# Data Models
class WebPubSubMessage(BaseModel):
    messageType: type_wps_message_type
    payload: Dict[str, Any]
    correlationId: Optional[str] = None

    @model_validator(mode="after")  # type: ignore
    def check_StartTest_payload(self):
        if self.messageType == "StartTest":
            if "groupId" not in self.payload:
                raise ValueError("StartTest requires groupId in payload!")

        return self

    @model_validator(mode="after")  # type: ignore
    def check_StartedTest_payload(self):
        if self.messageType == "StartedTest":
            if "testId" not in self.payload:
                raise ValueError("StartedTest requires testId in payload!")

        return self

    @model_validator(mode="after")  # type: ignore
    def check_Request_payload(self):
        if self.messageType == "Request":
            if self.correlationId is None:
                raise ValueError("Request requires correlationId to be valid!")

        return self


def wps_network_message_to_mg_message(
    message: OnGroupDataMessageArgs,
) -> WebPubSubMessage:
    return WebPubSubMessage(**dict(message.data))



# Standard library imports
from typing import Dict

# Third party imports
from rich.console import Console
from rich.live import Live
from rich.progress import Progress, SpinnerColumn, TaskID, TextColumn
from rich.table import Table

# Project imports
from mindgard.test import AttackState, Test, TestError


class TestUI:
    def __init__(self, test: Test, console: Console = Console()):
        self._console = console
        self._test = test

    def run(self) -> None:
        try:
            self._run()
        except TestError:
            # TODO: move error display to UI functions
            #       this is an interim step to stop the exception appearing twice.
            #       nb. we're only hiding errors we know to come from Test lib, not
            #       errors that might have come from the UI itself.
            pass

    def _run(self) -> None:
        test = self._test

        with test.state_wait_for(lambda state: state.submitting):
            with Progress(
                "{task.description}",
                SpinnerColumn(finished_text="[green3] Submitted!"),
                auto_refresh=True,
                console=self._console,
            ) as submit_progress:
                submit_task_id = submit_progress.add_task("Submitting...", start=True)

                with test.state_wait_for(lambda state: state.started):
                    submit_progress.update(submit_task_id, completed=100)

        ##################################
        # the test is in progress
        progress_table = Table.grid(expand=True)

        overall_task_progress = Progress()
        exceptions_progress = Progress("{task.description}")

        attack_progress = Progress(
            "{task.description}",
            SpinnerColumn(finished_text="done"),
            TextColumn("{task.fields[status]}"),
        )

        progress_table.add_row(overall_task_progress)
        progress_table.add_row(attack_progress)
        progress_table.add_row("")
        progress_table.add_row(exceptions_progress)

        def task_update(attack_progress: Progress, task: TaskID, attack: AttackState) -> None:
            if attack.state == "completed":
                if attack.errored:
                    attack_progress.update(task, status="[red3]failed", completed=1)
                else:
                    attack_progress.update(task, status="[chartreuse3]success", completed=1)
            elif attack.state == "running":
                attack_progress.update(task, status="[yellow]running", completed=0)
            else:
                attack_progress.update(task, status="[chartreuse1]queued", completed=0)

        with test.state_wait_for(lambda state: (len(state.attacks) > 0 or state.test_complete)) as state:
            attack_id_task_map: Dict[str, TaskID] = {}
            for attack in state.attacks:
                attack_id_task_map[attack.id] = attack_progress.add_task(f"Attack {attack.name}", total=1)
                task_update(attack_progress, attack_id_task_map[attack.id], attack)

            attacks_progress = overall_task_progress.add_task("Progress", total=len(state.attacks))
            completed_attacks = sum(1 for attack in state.attacks if attack.state == "completed")
            overall_task_progress.update(attacks_progress, completed=completed_attacks)

            model_exception_task_map: Dict[str, TaskID] = {}
            model_exception_counts: Dict[str, int] = {}
            for model_exception in state.model_exceptions:
                name = model_exception
                if model_exception_task_map.get(name) is None:
                    model_exception_task_map[name] = exceptions_progress.add_task("")
                    model_exception_counts[name] = 1
                else:
                    model_exception_counts[name] += 1

            for name, task in model_exception_task_map.items():
                exceptions_progress.update(
                    task, description=f"[dark_orange3][!!!] {name} x{model_exception_counts[name]}"
                )

        with Live(progress_table, refresh_per_second=10, console=self._console):
            while True:
                with test.state_then_wait_if(lambda state: not state.test_complete) as state:
                    for attack in state.attacks:
                        task = attack_id_task_map.get(attack.id)
                        if task is None:  # library could have changed the attack list
                            attack_id_task_map[attack.id] = attack_progress.add_task(f"Attack {attack.name}", total=1)
                            task = attack_id_task_map[attack.id]
                        task_update(attack_progress, task, attack)

                    completed_attacks = sum(1 for attack in state.attacks if attack.state == "completed")
                    overall_task_progress.update(attacks_progress, completed=completed_attacks)

                    model_exception_counts: Dict[str, int] = {}
                    for model_exception in state.model_exceptions:
                        name = model_exception
                        if model_exception_task_map.get(name) is None:
                            model_exception_task_map[name] = exceptions_progress.add_task("")
                        if model_exception_counts.get(name) is None:
                            model_exception_counts[name] = 1
                        else:
                            model_exception_counts[name] += 1
                    for name, task in model_exception_task_map.items():
                        exceptions_progress.update(
                            task, description=f"[dark_orange3][!!!] {name} x{model_exception_counts[name]}"
                        )

                    if state.test_complete:
                        break

        ##################################
        # the test is complete
        final_state = test.get_state()
        table = Table(title=f"Results - https://sandbox.mindgard.ai/r/test/{final_state.test_id}", width=80)
        table.add_column("Pass", style="cyan")
        table.add_column("Name", style="magenta")
        table.add_column("Risk", justify="right", style="green")

        for attack in final_state.attacks:
            if attack.errored:
                name = f"Error running '{attack.name}'"
                risk_str = "n/a"
                emoji = "❗️"
            else:
                name = attack.name
                risk_str = str(attack.risk)
                emoji = "✅️" if attack.passed else "❌‍"

            table.add_row(emoji, name, risk_str)

        self._console.print(table)



# Standard library imports
import textwrap

# Project imports
from mindgard.multiturn.attack import VALID_MULTI_TURNS

from .shared import add_shared_arguments
from .types import _SubparserType


def add_run_command(subparsers: _SubparserType):
    run_parser = subparsers.add_parser("run", help="Run a multiturn technique against your target system")
    run_subparser = run_parser.add_subparsers(title="Run subcommands", dest="run_subcommands", required=True)
    add_run_multiturn_command(run_subparser)
    return run_parser


def add_run_multiturn_command(run_subparser: _SubparserType):
    run_multiturn_parser = run_subparser.add_parser("multiturn", help="Run multiturn attacks")
    add_shared_arguments(run_multiturn_parser)
    run_multiturn_parser.add_argument(
        "--goal",
        required=True,
        type=str,
        help="The goal for all multi-turn techniques to run against your system",
    )
    run_multiturn_parser.add_argument(
        "--include",
        type=str,
        choices=VALID_MULTI_TURNS,
        help=textwrap.dedent(
            """
            Include a selected set of attacks in the test.
            """
        ),
        action="append",
        required=True,
    )
    return run_multiturn_parser



# Standard library imports
import uuid
from concurrent.futures import ThreadPoolExecutor
from time import sleep
from typing import Any

# Third party imports
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn

# Project imports
from mindgard.api.exceptions import ClientException
from mindgard.api.recon.guardrail import GetDetectResponse, ReconGuardrailClient
from mindgard.auth import load_access_token
from mindgard.constants import API_BASE
from mindgard.recon.guardrail_subcommand import GuardrailReconCommand
from mindgard.utils import CliResponse, request_user_input
from mindgard.wrappers.llm import LLMModelWrapper


def run_guardrail_recon(console: Console, model_wrapper: LLMModelWrapper, final_args: dict[str, Any]):
    access_token = load_access_token()
    guardrail_recon_service = ReconGuardrailClient(API_BASE, access_token, final_args["project_id"])
    guardrail_command = GuardrailReconCommand(guardrail_recon_service, model_wrapper)
    result: list[GetDetectResponse] = []

    def guardrail() -> None:
        recon_id = guardrail_command.start_detect(final_args["project_id"])
        guardrail_command.poll_detect(recon_id)
        detect_results = guardrail_command.fetch_recon_detect_result(recon_id)
        result.append(detect_results)

    def guardrail_fingerprint(recon_id: uuid.UUID) -> None:
        fp_recon_id = guardrail_command.start_fingerprint(recon_id)
        if fp_recon_id != recon_id:
            print(fp_recon_id, type(fp_recon_id))
            print(recon_id, type(recon_id))
            console.print(f"[red bold]Error: Fingerprinting session did not initialise correctly.[/red bold]")
            exit(CliResponse(1).code())
        guardrail_command.poll_fingerprint(fp_recon_id)
        fp_results = guardrail_command.fetch_recon_fingerprint_result(fp_recon_id)
        result.append(fp_results)

    def guardrail_with_spinner() -> None:
        with Progress(
            SpinnerColumn(style="bold yellow"),
            TextColumn("[bold yellow]Probing for guardrails…"),
            transient=True,
            console=console,
        ) as progress:
            task_id = progress.add_task("probing", total=None)

            with ThreadPoolExecutor(max_workers=1) as executor:
                guardrail_future = executor.submit(guardrail)

                while not guardrail_future.done():
                    sleep(0.15)

                progress.update(task_id, completed=100)
                guardrail_future.result()

    def fingerprint_with_spinner() -> None:
        """
        Start fingerprinting session worker(s) and display a spinner to the user
        """
        with Progress(
            SpinnerColumn(style="bold yellow"),
            TextColumn("[bold yellow]Fingerprinting guardrails…"),
            transient=True,
            console=console,
        ) as progress:
            task_id = progress.add_task("fingerprinting", total=None)

            with ThreadPoolExecutor(max_workers=1) as executor:
                guardrail_future = executor.submit(guardrail_fingerprint, result[0].id)

                while not guardrail_future.done():
                    sleep(0.15)

                progress.update(task_id, completed=100)
                guardrail_future.result()

    # Perform initial guardrail recon
    try:
        guardrail_with_spinner()
    except ClientException as ex:
        raise ex

    if len(result) == 0:
        console.print("Failed to get result from guardrail detection session")
        exit(CliResponse(1).code())

    console.print("\nProbing completed!", style="bold green")
    detected = result[0].result.guardrail_detected

    if not detected:
        # CASE: No guardrail detected
        console.print("\nNo clear signs of guardrail(s).", style="bold red")
        console.print(
            f"\n[underline]Reasoning[/underline]: The target didn't produce outputs that point to guardrail behavior."
        )
        console.print(
            f"[underline]Recommendation[/underline]: Run the complete mindgard test suite for open-ended probing to uncover any additional risks. See [yellow]mindgard test --help[/yellow] to learn more.\n"
        )
        exit(CliResponse(0).code())

    console.print("\nSome signs of guardrail(s) found.\n", style="bold blue")

    # Perform guardrail fingerprinting if guardrail detected
    guardrails_detected = None
    console.print(
        "Continue with fingerprinting to identify enabled guardrails?\nDisclaimer: this action will fire [underline]~120[/underline] prompts at the target system"
    )
    if request_user_input(console, "(y/n): ", ["y", "n"]) == "n":
        # CASE: Guardrail detected but user chose not to continue with fingerprinting
        console.rule(style="bold white")
        console.print(f"Guardrail reconnaissance results:")
        console.print("\nSome signs of guardrail(s) found.", style="bold blue")
        console.print(
            f"\n[underline]Reasoning[/underline]: The target produced outputs that point to guardrail presence."
        )
        console.print(
            f"[underline]Recommendation[/underline]: Shift from discovery to evasion. Run tests with different guardrail bypass domains using [yellow]mindgard test --domain bypass.\\[guardrail_name][yellow]\n"
        )
        exit(CliResponse(0).code())

    try:
        # Perform fingerprinting prompt request-response
        fingerprint_with_spinner()

        # Compile results
        guardrails_result = result[1]
        highest_confidence = max(getattr(x, "confidence", 0) for x in guardrails_result)
        guardrails_detected = [x for x in guardrails_result if getattr(x, "confidence", 0) == highest_confidence]
        names = " / ".join(getattr(x, "guardrail_pretty_name", "N/A") for x in guardrails_detected)

        console.rule(style="bold white")
        console.print(f"Guardrail reconnaissance results:\n")

        if highest_confidence == 0:
            # CASE: Guardrail responses matched no known fingerprints
            console.print(f"Response behaviour could not be matched to a known guardrail.", style="bold blue")
            console.print(
                f"\n[underline]Reasoning[/underline]: Responses to fingerprinting prompts did not match the patterns of known guardrails."
            )
            console.print(
                f"[underline]Recommendation[/underline]: We recommend re-running the recon process to try different prompts.\n"
            )
        else:
            # CASE: One or more guardrails matched known fingerprints
            console.print(f"The guardrail exhibits behaviours similar to {names}.", style="bold blue")
            console.print(
                f"\n[underline]Reasoning[/underline]: Custom inputs for {'these guardrails' if '/' in names else 'this guardrail'} were the most successful."
            )
            console.print(
                f"[underline]Recommendation[/underline]: Use the relevant Mindgard guardrail bypass domain to further validate and attempt evasion. Run [yellow]mindgard test --domain bypass.\\[guardrail_name][/yellow] with the selected domain.\n"
            )
        exit(CliResponse(0).code())

    except ClientException as ex:
        console.print(f"[red bold]Error: {ex.message}[/red bold]")
        exit(CliResponse(ex.status_code).code())



# Standard library imports
import csv
import logging
import time
from dataclasses import dataclass
from pathlib import Path

# Third party imports
import requests
from rich.console import Console

# Project imports
from mindgard.auth import require_auth
from mindgard.constants import API_BASE, EXIT_CODE_ERROR, VERSION
from mindgard.utils import print_to_stderr

console = Console()


class CreateDatasetError(Exception):
    pass


class InternalError(CreateDatasetError):
    pass


class UnauthorizedError(CreateDatasetError):
    pass


@dataclass
class CreateCustomDatasetRequest:
    seed_prompt: str
    perspective: str
    tone: str
    num_entries: int


@dataclass()
class CreateCustomDatasetCLIArgs:
    seed_prompt: str
    perspective: str
    tone: str
    output_filename: str
    num_entries: int


@require_auth
def create_custom_dataset(create_args: CreateCustomDatasetCLIArgs, access_token: str) -> None:
    """
    Create request to generate a custom dataset
    """
    url = f"{API_BASE}/dataset-generation"
    if Path(create_args.output_filename).exists():
        print_to_stderr(
            FileExistsError(
                f"File name: {create_args.output_filename} already exists. Please choose a different name by setting --output-filename <your_file_name>"
            )
        )
        exit(EXIT_CODE_ERROR)

    if create_args.num_entries < 1 or create_args.num_entries > 250:
        print_to_stderr(
            ValueError(
                f"Number of entries (--num_entries) must be between 1 and 250 (provided: {create_args.num_entries})"
            )
        )
        exit(EXIT_CODE_ERROR)

    try:

        json_payload = CreateCustomDatasetRequest(
            seed_prompt=create_args.seed_prompt,
            perspective=create_args.perspective,
            tone=create_args.tone,
            num_entries=create_args.num_entries,
        ).__dict__

        logging.debug(f"Dataset generation request payload: {json_payload}")
        response = requests.post(
            url=url,
            headers={
                "Authorization": f"Bearer {access_token}",
                "User-Agent": f"mindgard-cli/{VERSION}",
                "X-User-Agent": f"mindgard-cli/{VERSION}",
                "Content-Type": "application/json",
            },
            json=json_payload,
        )
        if response.status_code == 201:
            with console.status("Dataset generation in progress...", spinner="dots"):
                logging.debug(f"Dataset generation started with request ID: {response.json()['id']}")
                while True:
                    time.sleep(0.5)
                    datasets = get_generated_dataset(
                        dataset_id=response.json()["id"], file_name=create_args.output_filename
                    )
                    if datasets and datasets.get("is_complete", False):
                        console.print(f"Dataset generation completed. Filename is {create_args.output_filename}")
                        break
        response.raise_for_status()
    except requests.HTTPError as e:
        if e.response.status_code == 401:
            raise UnauthorizedError("Invalid access token") from e
        elif e.response.status_code == 403:
            raise UnauthorizedError("Access denied") from e
        else:
            raise InternalError("An unexpected error occurred while creating dataset generation request") from e
    except Exception as e:
        raise InternalError("An unexpected error occurred creating dataset generation request") from e


@require_auth
def get_generated_dataset(dataset_id: str, access_token: str, file_name: str) -> None:
    """
    Get the generated dataset
    """
    url = f"{API_BASE}/dataset-generation/{dataset_id}"

    try:
        response = requests.get(
            url=url,
            headers={
                "Authorization": f"Bearer {access_token}",
                "User-Agent": f"mindgard-cli/{VERSION}",
                "X-User-Agent": f"mindgard-cli/{VERSION}",
                "Content-Type": "application/json",
            },
        )
        if response.status_code == 200:
            with open(f"{file_name}", "w", newline="") as f:
                writer = csv.writer(f)
                for item in response.json()["items"]:
                    writer.writerow([item])

        return response.json()
    except requests.HTTPError as e:
        if e.response.status_code == 401:
            raise UnauthorizedError("Invalid access token") from e
        elif e.response.status_code == 403:
            raise UnauthorizedError("Access denied") from e
        else:
            raise InternalError("An unexpected error occurred while fetching generated dataset") from e
    except Exception as e:
        raise InternalError("An unexpected error occurred fetching generated dataset") from e



# Standard library imports
import uuid

# Third party imports
from pydantic import BaseModel


class StartReconRequest(BaseModel):
    project_id: str


class StartReconResponse(BaseModel):
    recon_id: uuid.UUID


class GetReconRequest(BaseModel):
    recon_id: uuid.UUID



# Standard library imports
from dataclasses import dataclass
from typing import Any, Callable, Dict, Literal, Optional, Tuple, get_args

# Third party imports
import requests
from rich.progress import Progress, TaskID
from rich.table import Table


@dataclass
class ExceptionCountTuple:
    task_id: TaskID
    count: int


# Web pub sub message types
type_wps_message_type = Literal["StartTest", "StartedTest", "Response", "Request"]

# UI task map stores a name for a task along with its taskid to be updated in progress object
type_ui_task_map = Dict[str, TaskID]
# Stores an exception message with the task to update, and the number of times the exception has been thrown
type_ui_exception_map = Dict[str, ExceptionCountTuple]

# interfaces that submit, polling and output functions MUST MEET to be usable in run_poll_display.
type_submit_func = Callable[[str, type_ui_exception_map, Progress], Any]
type_polling_func = Callable[
    [str, Any, type_ui_task_map, Progress],
    Optional[Any],
]
type_output_func = Callable[[Any, bool], Optional[Table]]

# Orchestrator submit/request enums
type_orchestrator_attack_pack = Literal["sandbox", "threat_intel"]
type_orchestrator_source = Literal["threat_intel", "user", "mindgard"]
type_model_presets = Literal[
    "huggingface-openai",
    "openai-compatible",
    "huggingface",
    "openai",
    "azure-openai",
    "azure-aistudio",
    "anthropic",
    "tester",
    "custom",
]
type_model_presets_list: Tuple[type_model_presets, ...] = get_args(type_model_presets)

# Types for dependency injecting get/post request functions into over api_post and api_get in orchestrator
type_post_request_function = Callable[[str, str, Dict[str, Any]], requests.Response]
type_get_request_function = Callable[[str, str], requests.Response]

# Different log levels
log_levels = [
    "critical",
    "fatal",
    "error",
    "warn",
    "warning",
    "info",
    "debug",
    "notset",
]  # [n.lower() for n in logging.getLevelNamesMapping().keys()]

valid_llm_datasets = {
    "customerservice": "BadCustomer",
    "finance": "BadFinance",
    "legal": "BadLegal",
    "medical": "BadMedical",
    "injection": "SqlInjection",
    "rce": "Xss",
    "xss": "Xss",
    "bypass.deepset": "GuardrailBypassDeepset",
    "bypass.llmguard": "GuardrailBypassLlmGuard",
    "bypass.metapromptguardv1": "GuardrailBypassMetaPromptGuardV1",
    "bypass.protectaiv1": "GuardrailBypassProtectAiV1",
    "bypass.protectaiv2": "GuardrailBypassProtectAiV2",
    "bypass.vijilpromptinjection": "GuardrailBypassVijilPromptInjection",
    "toxicity": "Toxicity",
    "toxicity.discrimination": "ToxicityDiscrimination",
    "toxicity.harassment": "ToxicityHarassment",
    "toxicity.hate_speech": "ToxicityHateSpeech",
    "toxicity.profanity": "ToxicityProfanity",
    "information_disorder": "InformationDisorder",
    "information_disorder.disinformation": "InformationDisorderDisinformation",
    "information_disorder.misinformation": "InformationDisorderMisinformation",
    "cybersecurity": "Cybersecurity",
    "cybersecurity.cloud_attacks": "CybersecurityCloudAttacks",
    "cybersecurity.control_system_attacks": "CybersecurityControlSystemAttacks",
    "cybersecurity.cryptographic_attacks": "CybersecurityCryptographicAttacks",
    "cybersecurity.cyber_crime": "CybersecurityCyberCrime",
    "cybersecurity.evasion_techniques": "CybersecurityEvasionTechniques",
    "cybersecurity.hardware_attacks": "CybersecurityHardwareAttacks",
    "cybersecurity.intrusion_techniques": "CybersecurityIntrusionTechniques",
    "cybersecurity.iot_attacks": "CybersecurityIotAttacks",
    "cybersecurity.malware_attacks": "CybersecurityMalwareAttacks",
    "cybersecurity.network_attacks": "CybersecurityNetworkAttacks",
    "cybersecurity.web_application_attacks": "CybersecurityWebApplicationAttacks",
    "business_risk": "BusinessRisk",
    "business_risk.copyright": "BusinessRiskCopyright",
    "business_risk.pii": "BusinessRiskPii",
    "harmful": "Harmful",
    "harmful.dangerous_content": "HarmfulDangerousContent",
    "harmful.illegal": "HarmfulIllegal",
    "harmful.self_harm": "HarmfulSelfHarm",
    "harmful.sexually_explicit": "HarmfulSexuallyExplicit",
    "harmful.violence": "HarmfulViolence",
}



# Standard library imports
import logging
import uuid
from typing import Optional
from urllib.parse import urljoin

# Third party imports
from pydantic import BaseModel

from ..exceptions import ClientException
from .recon_prompt_events import ReconPromptEventsClient


class StartDetectRequest(BaseModel):
    project_id: str


class StartDetectResponse(BaseModel):
    recon_id: uuid.UUID


class StartFingerprintRequest(BaseModel):
    recon_id: uuid.UUID


class StartFingerprintResponse(BaseModel):
    recon_id: uuid.UUID


class GetReconRequest(BaseModel):
    recon_id: uuid.UUID


class ReconResult(BaseModel):
    guardrail_detected: bool
    detected_guardrails: list[str] = []


class GetDetectResponse(BaseModel):
    id: uuid.UUID
    state: str
    result: Optional[ReconResult] = None
    reason: Optional[str] = None
    recommendation: Optional[str] = None
    project_id: str


class GetFingerprintResponse(BaseModel):
    guardrail_name: str
    guardrail_pretty_name: str
    confidence: float
    errors: int


class ReconGuardrailClient(ReconPromptEventsClient):
    def __init__(self, base_url: str, access_token: str, project_id: str):
        super().__init__(base_url, access_token, project_id)
        self._reconn_detection_url = urljoin(base_url + "/", f"recon/guardrail/detection")
        self._reconn_fingerprint_url = urljoin(base_url + "/", f"recon/guardrail/fingerprint")
        self._project_id = project_id

    def start_detect(self, request: StartDetectRequest) -> StartDetectResponse:
        """
        Start the detection process, will start a new reconnaissance session

        Args:
            request (StartDetectRequest): Info connecting this operation to a target name

        Raises:
            ClientException: Raised if the Mindgard service returns an error

        Returns:
            StartDetectResponse: Mindgard service response from the start request
        """
        response = self._session.post(
            self._reconn_detection_url, params={"project_id": self._project_id}, data=request.model_dump_json()
        )

        if response.status_code != 201:
            logging.debug(f"Failed to start recon: {response.json()} - {response.status_code}")
            raise ClientException(response.json(), response.status_code)
        try:
            return StartDetectResponse.model_validate(response.json())
        except Exception as e:
            raise ClientException("Problem parsing StartDetectResponse", response.status_code) from e

    def start_fingerprint(self, request: StartFingerprintRequest) -> StartFingerprintResponse:
        """
        Start the fingerprinting process for an existing reconnaissance session

        Args:
            request (StartFingerprintRequest): Info connecting this operation to an existing recon session

        Raises:
            ClientException: Raised if the Mindgard service returns an error

        Returns:
            StartFingerprintResponse: Mindgard service response from the start request
        """
        response = self._session.post(
            self._reconn_fingerprint_url, params={"project_id": self._project_id}, data=request.model_dump_json()
        )

        if response.status_code != 201:
            logging.debug(f"Failed to start fingerprinting: {response.json()} - {response.status_code}")
            raise ClientException(response.json(), response.status_code)
        try:
            return StartFingerprintResponse.model_validate(response.json())
        except Exception as e:
            raise ClientException("Problem parsing StartFingerprintResponse", response.status_code) from e

    def get_detect_result(self, request: GetReconRequest) -> Optional[GetDetectResponse]:
        """
        Query the Mindgard service for the results of a guardrail detect recon operation

        Args:
            request (GetReconRequest): Request object containing recon_id

        Raises:
            ClientException: Raised if there is a service error fetching the detection results

        Returns:
            Optional[GetDetectResponse], None]: Result for guardrail detection, or None if no results returned
        """
        response = self._session.get(
            self._reconn_detection_url, params={"recon_id": str(request.recon_id), "project_id": self._project_id}
        )

        if response.status_code == 404:
            logging.debug(f"No reconn found for source_id={request.recon_id}: {response.text}")
            return None

        if response.status_code != 200:
            logging.debug(f"Failed to get recon result: {response.text} - {response.status_code}")
            raise ClientException(response.json(), response.status_code)
        try:
            return GetDetectResponse.model_validate(response.json())
        except Exception as e:
            raise ClientException("Problem parsing GetDetectResponse", response.status_code) from e

    def get_fingerprint_result(self, request: GetReconRequest) -> Optional[list[GetFingerprintResponse]]:
        """
        Query the Mindgard service for the results of a guardrail fingerprinting recon operation

        Args:
            request (GetReconRequest): Request object containing recon_id

        Raises:
            ClientException: Raised if there is a service error fetching the fingerprint results

        Returns:
            Optional[list[GetFingerprintResponse]]: Results for each guardrail fingerprinted, or None if no results returned
        """
        response = self._session.get(
            self._reconn_fingerprint_url, params={"recon_id": str(request.recon_id), "project_id": self._project_id}
        )

        if response.status_code == 404:
            logging.debug(f"No reconn found for source_id={request.recon_id}: {response.text}")
            return None

        if response.status_code != 200:
            logging.debug(f"Failed to get recon result: {response.text} - {response.status_code}")
            raise ClientException(response.json(), response.status_code)

        resp_json = response.json()

        if not isinstance(resp_json, list):
            logging.debug(f"Expected list in fingerprint result, got: {type(resp_json)}")
            return None
        try:
            return [GetFingerprintResponse.model_validate(item) for item in resp_json]
        except Exception as e:
            raise ClientException("Problem parsing GetFingerprintResponse", response.status_code) from e



# Standard library imports
from argparse import ArgumentParser, _SubParsersAction
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    _SubparserType = _SubParsersAction[ArgumentParser]
else:
    _SubparserType = Any



# Standard library imports
import json
import os
from dataclasses import dataclass
from typing import Any, Dict


@dataclass
class CliConfig:
    AUTH0_DOMAIN: str
    AUTH0_CLIENT_ID: str
    AUTH0_AUDIENCE: str
    API_BASE: str
    DASHBOARD_URL: str


def get_config_directory() -> str:
    config_dir = os.environ.get("MINDGARD_CONFIG_DIR")
    return config_dir or os.path.join(os.path.expanduser("~"), ".mindgard")


def create_config_directory() -> None:
    os.makedirs(get_config_directory(), exist_ok=True)


def get_token_file() -> str:
    return os.path.join(get_config_directory(), "token.txt")


def get_instance_file() -> str:
    return os.path.join(get_config_directory(), "instance.txt")


def is_instance_set() -> bool:
    return os.path.isfile(os.path.join(get_config_directory(), "instance.txt"))


def instance_auth_config(config: Dict[str, Any]) -> CliConfig:
    return CliConfig(
        AUTH0_DOMAIN=os.getenv("AUTH0_DOMAIN", str(config["domain"])),
        AUTH0_CLIENT_ID=os.getenv("AUTH0_CLIENT_ID", str(config["clientId"])),
        AUTH0_AUDIENCE=os.getenv("AUTH0_AUDIENCE", str(config["audience"])),
        API_BASE=os.getenv("API_BASE", str(config["apiBase"])),
        DASHBOARD_URL=os.getenv("DASHBOARD_URL", str(config["dashboardUrl"])),
    )


def sandbox_auth_config() -> CliConfig:
    return CliConfig(
        AUTH0_DOMAIN=os.getenv("AUTH0_DOMAIN", "login.sandbox.mindgard.ai"),
        AUTH0_CLIENT_ID=os.getenv("AUTH0_CLIENT_ID", "U0OT7yZLJ4GEyabar11BENeQduu4MaNO"),
        AUTH0_AUDIENCE=os.getenv("AUTH0_AUDIENCE", "https://marketplace-orchestrator.com"),
        API_BASE=os.getenv("API_BASE", "https://api.sandbox.mindgard.ai/api/v1"),
        DASHBOARD_URL=os.getenv("DASHBOARD_URL", "https://sandbox.mindgard.ai"),
    )


def load_auth_config() -> CliConfig:
    if is_instance_set():
        try:
            # If an instance config exists then load the variables
            with open(get_instance_file(), "r") as f:
                instance_config = json.load(f)

            return instance_auth_config(instance_config)
        except:
            # jsonDecodeError & KeyError
            # TODO: Handle this automatically
            raise KeyError(
                "Unable to load instance configuration. To create a new configuration run `mindgard logout` and then `mindgard login --instance XXXX`"
            )
    else:
        # Default sandbox constants
        return sandbox_auth_config()



# Standard library imports
import logging
import uuid
from typing import Literal, Optional
from urllib.parse import urljoin

# Third party imports
import requests
from pydantic import BaseModel

# Project imports
from mindgard.version import VERSION

from ..exceptions import ClientException

ALLOWED_EVENT_TYPES = Literal["prompt_request", "complete"]
# Allowed Event Subjects was originally used for validating the allowed types in
# a Pop or Push request. These types are now only set for GUARDRAIL recons.
# Subsequently, most of the values in this list will never be validated against.
ALLOWED_EVENT_SUBJECTS = Literal[
    "GUARDRAIL_DETECT",
    "GUARDRAIL_FINGERPRINT",
    "INPUT_ENCODING",
    "OUTPUT_ENCODING",
    "NON_CONTEXTUAL",
    "OUTPUT_FORMATTING",
    "OUTPUT_RENDERING",
    "CODE_GENERATION",
    "SYSTEM_PROMPT_EXTRACTION",
]
RESULT_EVENT_TYPE = Literal["prompt_result"]


class PopEventRequest(BaseModel):
    source_id: uuid.UUID
    event_type: list[ALLOWED_EVENT_TYPES]
    event_subject: Optional[ALLOWED_EVENT_SUBJECTS] = None


class PromptRequest(BaseModel):
    id: Optional[str] = None
    prompt: str
    language: str


class PopEventResponse(BaseModel):
    event_id: str
    event_type: str
    source_id: uuid.UUID
    prompt_request: Optional[list[PromptRequest]] = None


class PromptResult(BaseModel):
    id: Optional[str] = None
    content: Optional[str] = None
    duration_ms: Optional[float] = None
    prompt_request: PromptRequest
    error_message: Optional[str] = None
    error_code: Optional[int] = None


class PushEventRequest(BaseModel):
    source_id: uuid.UUID
    event_type: RESULT_EVENT_TYPE
    prompt_result: list[PromptResult]
    event_subject: Optional[ALLOWED_EVENT_SUBJECTS] = None


class PushEventResponse(BaseModel):
    event_id: str


class ReconPromptEventsClient:
    def __init__(self, base_url: str, access_token: str, project_id: str):
        self._base_url = base_url
        self._pop_events_url = urljoin(self._base_url + "/", f"events/prompt_request_response/pop")
        self._push_events_url = urljoin(self._base_url + "/", f"events/prompt_request_response/push")
        self._access_token = access_token
        self._session = requests.Session()
        self._project_id = project_id
        self._session.headers.update(
            {
                "Authorization": f"Bearer {self._access_token}",
                "Content-Type": "application/json",
                "User-Agent": f"mindgard-cli/{VERSION}",
                "X-User-Agent": f"mindgard-cli/{VERSION}",
            }
        )

    def pop(self, request: PopEventRequest) -> Optional[PopEventResponse]:
        response = self._session.post(
            self._pop_events_url,
            params={"project_id": self._project_id},
            data=request.model_dump_json(),
        )

        if response.status_code == 404:
            logging.debug(f"No event found for source_id={request.source_id}: {response.text}")
            return None

        if response.status_code != 200:
            logging.debug(f"Failed to get recon events: {response.text} - {response.status_code}")
            raise ClientException(response.json(), response.status_code)

        try:
            return PopEventResponse.model_validate(response.json())
        except Exception as e:
            raise ClientException("Problem parsing PopEventResponse", response.status_code) from e

    def push(self, request: PushEventRequest) -> PushEventResponse:
        logging.debug(f"pushing event: {request}")
        for i in range(3):
            try:
                response = self._session.post(
                    self._push_events_url,
                    params={"project_id": self._project_id},
                    data=request.model_dump_json(),
                )
                break
            except requests.exceptions.ConnectionError:
                logging.debug(f"Connection error on try {i}! Trying again...")

        if response.status_code != 201:
            response_body = response.text
            logging.debug(
                f"Retrying with sanitized results as failed to get prompt results: {response_body} - {response.status_code}"
            )
            for result in request.prompt_result:
                result.content = ""
                result.error_code = -1
                result.error_message = "Failed to contact Mindgard"
            response = self._session.post(self._push_events_url, data=request.model_dump_json())

            if response.status_code != 201:
                logging.debug(
                    f"Retrying with sanitized results failed to get prompt results: {response.text} - {response.status_code}"
                )
                raise ClientException(response.text, response.status_code)

        logging.debug(f"Got prompt results: {response.json()} - {response.status_code}")
        return PushEventResponse.model_validate(response.json())



from .types import _SubparserType


def add_login_command(subparsers: _SubparserType):
    login_parser = subparsers.add_parser("login", help="Login to the Mindgard platform")
    login_parser.add_argument(
        "--instance",
        nargs="?",
        type=str,
        help="Point to your deployed Mindgard instance. If not provided, cli will point towards Mindgard Sandbox",
    )
    return login_parser



# Standard library imports
import logging
import uuid
from typing import Any
from urllib.parse import urljoin

# Third party imports
from pydantic import BaseModel, Field

from .attack_prompt_events import AttackPromptEventsClient
from .exceptions import ClientException


class StartRequest(BaseModel):
    """
    Request model for starting a multiturn attack session.

    Attributes:
        project_id (str): The name of the target for the attack.
        goal (str): The goal or objective of the attack session.
        attack_name (str): The name of the attack to run.
    """

    project_id: str
    goal: str
    attack_name: str


class StartResponse(BaseModel):
    """
    Response model for a multiturn attack start request.

    Attributes:
        run_id (uuid.UUID): Unique identifier for the started run/session.
        state (str): The current state of the session.
    """

    run_id: uuid.UUID
    state: str


class ResultsRequest(BaseModel):
    """
    Request model for retrieving results of a multiturn attack session.

    Attributes:
        run_id (uuid.UUID): Unique identifier for the started run/session.
    """

    run_id: uuid.UUID


class ResultsResponse(BaseModel):
    """
    Response model for results of a multiturn attack session.

    Attributes:
        id (str): Unique identifier for the result.
        state (str): The current state of the session.
        project_id (str): The project ID.
        result (Any): The result data, if available.
    """

    state: str
    attack_session_ids: list[uuid.UUID] = Field(default_factory=list)


class AttackMultiturnClient(AttackPromptEventsClient):
    """
    Client for managing multiturn attack sessions via the API.

    Inherits from AttackPromptEventsClient and provides methods to start and interact
    with multiturn attack sessions, handling authentication and request formatting.

    Args:
        base_url (str): The base URL of the API endpoint.
        access_token (str): Bearer token for authentication.
    """

    def __init__(self, base_url: str, access_token: str, project_id: str):
        """
        Initialize the AttackMultiturnClient.

        Args:
            base_url (str): The base URL of the API endpoint.
            access_token (str): Bearer token for authentication.
        """
        super().__init__(base_url, access_token)
        self._attack_url = urljoin(base_url + "/", f"user-attack-sessions")
        self._project_id = project_id

    def start(self, req: StartRequest) -> StartResponse:
        """
        Start a new multiturn attack session.

        Args:
            req (StartRequest): The request containing target and goal information.

        Returns:
            StartResponse: The response containing the run ID and session state.

        Raises:
            ClientException: If the request fails or response validation fails.
        """
        response = self._session.post(
            self._attack_url, params={"project_id": self._project_id}, data=req.model_dump_json()
        )

        if response.status_code != 201:
            logging.debug(f"Failed to start multiturn attack: {response.json()} - {response.status_code}")
            raise ClientException(response.text, response.status_code)

        try:
            return StartResponse.model_validate(response.json())
        except Exception as e:
            raise ClientException("Problem parsing StartResponse", response.status_code) from e

    def get_results(self, req: ResultsRequest) -> ResultsResponse:
        response = self._session.get(self._attack_url + f"/{req.run_id}", params={"project_id": self._project_id})

        if response.status_code != 200:
            logging.debug(f"Failed to get multiturn attack: {response.json()} - {response.status_code}")
            raise ClientException(response.text, response.status_code)

        try:
            return ResultsResponse.model_validate(response.json())
        except Exception as e:
            raise ClientException("Problem parsing ResultsResponse", response.status_code) from e



"""
Temporary module to split off the __main__ for test execution using
new library system.
"""

# Standard library imports
from threading import Thread
from typing import Any, Dict

# Third party imports
import requests

# Project imports
from mindgard.auth import clear_token, load_access_token
from mindgard.constants import API_BASE, EXIT_CODE_ERROR, EXIT_CODE_NOT_PASSED, EXIT_CODE_PASSED
from mindgard.test import ForbiddenError, LLMModelConfig, Test, TestConfig, UnauthorizedError
from mindgard.test_ui import TestUI
from mindgard.utils import print_to_stderr
from mindgard.wrappers.llm import LLMModelWrapper


def run_test(final_args: Dict[str, Any], model_wrapper: LLMModelWrapper):
    access_token = load_access_token()
    if not access_token:
        print_to_stderr("\033[1;37mRun `mindgard login`\033[0;0m to authenticate.")
        exit(EXIT_CODE_ERROR)

    test_config = TestConfig(
        api_base=API_BASE,
        api_access_token=access_token,
        project_id=final_args["project_id"],
        attack_source="user",
        dataset_domain=final_args.get("domain", None),
        parallelism=final_args["parallelism"],
        model=LLMModelConfig(
            wrapper=model_wrapper,
            system_prompt=final_args["system_prompt"],
        ),
        exclude=final_args.get("exclude", None),
        include=final_args.get("include", None),
    )
    if attack_pack := final_args.get("attack_pack"):
        test_config.attack_pack = attack_pack

    test = Test(test_config)
    test_ui = TestUI(test)

    # daemonize to prevent blocking the main thread's exit if test.run raises exception
    test_ui_thread = Thread(target=test_ui.run, name="TestUI")
    test_ui_thread.start()

    try:
        test.run()
    except ForbiddenError:
        exit(EXIT_CODE_ERROR)
    except UnauthorizedError:
        print_to_stderr("Access token is invalid. Please re-authenticate using `mindgard login`")
        clear_token()
        exit(EXIT_CODE_ERROR)
    except requests.HTTPError as e:
        if "Unauthorized" in str(e):
            print_to_stderr("Access token is invalid. Please re-authenticate using `mindgard login`")
            clear_token()
            exit(EXIT_CODE_ERROR)
        else:
            raise e

    test_ui_thread.join()

    if test.get_state().passed:
        exit(EXIT_CODE_PASSED)
    elif test.get_state().passed == False:
        exit(EXIT_CODE_NOT_PASSED)



from ..constants import ENABLE_DEV_FEATURES
from .shared import add_shared_arguments
from .types import _SubparserType


def add_recon_command(subparsers: _SubparserType):
    recon_parser = subparsers.add_parser(
        "recon", help="Run various reconnaissance techniques against your target system"
    )
    recon_subparsers = recon_parser.add_subparsers(title="Recon subcommands", dest="recon_subcommand", required=True)

    # Define prod recon subcommands
    subcommands = {
        "guardrail": "Run guardrail reconnaissance against your target system",
        "input-encoding": "Run input encoding reconnaissance against your target system",
        "output-encoding": "Run output encoding reconnaissance against your target system",
        "system-prompt-extraction": "Run system prompt extraction against your target system",
        "tool-discovery": "Run tool discovery reconnaissance against your target system",
    }
    # Add dev recon subcommands
    if ENABLE_DEV_FEATURES:
        subcommands["non-contextual"] = "Run non-contextual reconnaissance against your target system"
        subcommands["output-formatting"] = "Run output formatting reconnaissance against your target system"
        subcommands["output-rendering"] = "Run output rendering reconnaissance against your target system"
        subcommands["code-generation"] = "Run code generation reconnaissance against your target system"

    for cmd, help_text in subcommands.items():
        subparser = recon_subparsers.add_parser(cmd, help=help_text)
        add_shared_arguments(subparser)

    return recon_parser



# mindgard/cli_args/__init__.py
# Standard library imports
import argparse
from pathlib import Path
from typing import TYPE_CHECKING, Any, List

from ..constants import ENABLE_DEV_FEATURES, VERSION
from ..types import log_levels
from .create import add_create_command
from .list import add_list_command
from .login import add_login_command
from .logout import add_logout_command
from .recon import add_recon_command
from .run import add_run_command
from .sandbox import add_sandbox_command
from .test import add_test_command
from .validate import add_validate_command


def create_parser():
    parser = argparse.ArgumentParser(
        description="Securing AIs",
        prog="mindgard",
        usage="%(prog)s [command] [options]",
        epilog="Enjoy the program! :)",
        add_help=True,
    )
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {VERSION}", help="Show the current version number"
    )
    parser.add_argument(
        "--log-level",
        type=str,
        help="Specify the output verbosity",
        choices=log_levels,
        required=False,
        default="warn",
    )

    parser.add_argument(
        "--log-file",
        type=Path,
        help="path to log file",
        required=False,
        default=None,
    )

    subparsers = parser.add_subparsers(
        dest="command", title="commands", description="Use these commands to interact with the Mindgard API"
    )
    # Add prod commands
    add_login_command(subparsers)
    add_logout_command(subparsers)
    add_sandbox_command(subparsers)
    add_list_command(subparsers)
    add_test_command(subparsers)
    add_validate_command(subparsers)
    add_recon_command(subparsers)
    add_create_command(subparsers)
    add_run_command(subparsers)

    # Add dev commands
    if ENABLE_DEV_FEATURES:
        pass

    return parser


def parse_args(args: List[str]) -> argparse.Namespace:
    parser = create_parser()
    return parser.parse_args(args)



# Standard library imports
import logging
import time
from typing import Callable, Optional

# Third party imports
from azure.messaging.webpubsubclient import WebPubSubClient
from azure.messaging.webpubsubclient.models import OnGroupDataMessageArgs
from rich.progress import Progress

# Project imports
from mindgard.exceptions import (
    ErrorCode,
    MGException,
    handle_exception_callback,
)
from mindgard.types import (
    ExceptionCountTuple,
    type_ui_exception_map,
)
from mindgard.wrappers.llm import ContextManager, LLMModelWrapper, PromptResponse


def llm_message_handler(
    ui_exception_map: type_ui_exception_map,
    ui_exception_progress: Progress,
    submitted_test_id_update: Callable[[str], None],
    model_wrapper: LLMModelWrapper,
    ws_client: WebPubSubClient,
) -> Callable[[OnGroupDataMessageArgs], None]:
    def _handle_visual_exception_callback(text: str) -> None:
        if not ui_exception_map.get(text, None):
            ui_exception_map[text] = ExceptionCountTuple(ui_exception_progress.add_task(""), 0)

        ui_exception_map[text].count += 1
        ui_exception_progress.update(
            ui_exception_map[text].task_id,
            description=f"[dark_orange3][!!!] {text} (x{ui_exception_map[text].count})",
        )

    def temp_handler(e: Exception) -> ErrorCode:
        return handle_exception_callback(e, _handle_visual_exception_callback)

    context_manager = ContextManager()

    def recv_message_handler(msg: OnGroupDataMessageArgs) -> None:
        if msg.data["messageType"] == "Request":
            logging.debug(f"received request {msg.data=}")
            context_id = msg.data["payload"].get("context_id", None)
            context = context_manager.get_context_or_none(context_id)
            content = msg.data["payload"]["prompt"]

            response: str = ""
            error_code: Optional[ErrorCode] = None
            duration_ms: float = 0.0

            try:
                # would pose being explicit with __call__ so we can ctrl+f easier, not a very clear shorthand
                prompt_response = model_wrapper.__call__(
                    content=content,
                    with_context=context,
                )
                response = prompt_response.response
                duration_ms = prompt_response.duration_ms
            except MGException as mge:
                error_code = temp_handler(mge)
                if error_code == "CLIError":
                    raise mge
                else:  # ensure we have access to the eexception trace in debug logging before continuing
                    logging.debug(f"error code: {error_code}", exc_info=True)
            except Exception as e:
                raise e
            finally:
                # we always try to send a response
                replyData = {
                    "correlationId": msg.data["correlationId"],
                    "messageType": "Response",
                    "status": "ok",  # Deprecated but kept for compatibility
                    "payload": {"response": response, "error": error_code, "duration_ms": duration_ms},
                }
                logging.debug(f"sending response {replyData=}")

                ws_client.send_to_group("orchestrator", replyData, data_type="json")  # type: ignore
        elif (
            msg.data["messageType"] == "StartedTest"
        ):  # should be something like "Submitted", upstream change required.
            submitted_test_id_update(msg.data["payload"]["testId"])
        else:
            pass

    return recv_message_handler



# Standard library imports
import logging
import time
import uuid
from typing import Optional

from ..api import attack_multiturn, attack_prompt_events
from ..api.exceptions import ClientException
from ..exceptions import MGException
from ..wrappers.llm import Context, LLMModelWrapper, PromptResponse

logger = logging.getLogger(__name__)

VALID_MULTI_TURNS = ("actor_attack", "crescendo", "skeleton_key")


class MultiturnAttack:
    def __init__(
        self,
        call_system_under_test: LLMModelWrapper,
        ma_client: attack_multiturn.AttackMultiturnClient,
        project_id: str,
        goal: str,
        attack_name: str,
        context: Optional[Context] = None,
    ):
        self._attack_name = attack_name
        self._goal = goal
        self._project_id = project_id
        self._maClient = ma_client
        self._call_system_under_test = call_system_under_test
        self._context = context or Context()

    def start(self) -> attack_multiturn.StartResponse:
        logger.debug(f"Starting multi-turn {self._attack_name} for project: {self._project_id}")
        return self._maClient.start(
            attack_multiturn.StartRequest(project_id=self._project_id, goal=self._goal, attack_name=self._attack_name)
        )

    def poll(self, request_uuid: uuid.UUID) -> None:

        completed = False
        event: Optional[attack_prompt_events.Event] = None

        logger.debug(f"Polling for prompt requests to process: {request_uuid}")
        while not completed:
            rt = time.monotonic()
            event, completed = self.__poll_inner(request_uuid, context=self._context)
            if completed:
                break
            # sleep if last request < 0.5s
            wait_time = max(0, (rt + 0.5) - time.monotonic())
            time.sleep(wait_time)

        logger.debug(f"All prompt requests processed for id: {request_uuid}")
        logger.debug(f"Fetching multi-turn {self._attack_name} result...")

    def __poll_inner(
        self, request_uuid: uuid.UUID, context: Context
    ) -> tuple[Optional[attack_prompt_events.Event], bool]:
        request = attack_prompt_events.PopRequest(runId=request_uuid)

        event = self._maClient.pop(request)
        if event is None:
            return None, False

        if event.runStatus == "COMPLETED":
            logger.debug(f"Attack completed for id: {request_uuid}")
            return event, True

        logger.debug(f"Processing prompt request")
        response: Optional[PromptResponse] = None
        error_code = None
        error_message = None
        duration_ms = 0
        content = ""
        try:
            response = self._call_system_under_test(event.event.query, with_context=context)
            content = response.response
            duration_ms = response.duration_ms
            logger.debug(f"{content=}")
        except MGException as e:
            logger.debug(f"Error communicating with target application: {e}")
            error_code = getattr(e, "status_code", None)
            error_message = str(e)
            logger.debug(f"{error_code=} {error_message=}")

        logger.debug("Pushing prompt result for request")
        result_push_request = attack_prompt_events.ResultPushRequest(
            requestEventID=event.event.id,
            sessionID=event.event.sessionID,
            referrerID=event.event.referrerID,
            attackName=event.event.attackName,
            turnIndex=event.event.turnIndex,
            durationMs=duration_ms,
            response=content,
            errorCode=error_code,
            errorMessage=error_message,
        )
        try:
            self._maClient.push(result_push_request)
        except ClientException:
            logger.debug("Failed to push response from target")
            result_push_request.response = ""
            result_push_request.errorCode = -1
            result_push_request.errorMessage = "Failed to contact Mindgard"
            self._maClient.push(result_push_request)

        return event, False

    def get_results(self, run_uuid: uuid.UUID) -> attack_multiturn.ResultsResponse:
        return self._maClient.get_results(attack_multiturn.ResultsRequest(run_id=run_uuid))



# Standard library imports
import argparse

from ..types import type_model_presets_list


def add_shared_arguments(parser: argparse.ArgumentParser):
    parser.add_argument("target", nargs="?", type=str, help="This is your own model identifier.")
    parser.add_argument(
        "--config-file", type=str, help="Path to mindgard.toml config file", default=None, required=False
    )
    parser.add_argument(
        "--json", action="store_true", help="Output the info in JSON format.", required=False, default=False
    )
    parser.add_argument("--headers", type=str, help="The headers to use. Comma separated list.", required=False)
    parser.add_argument(
        "--header", type=str, help="The headers to use, repeat flag for each header.", action="append", required=False
    )
    parser.add_argument("--preset", type=str, help="The preset to use", choices=type_model_presets_list, required=False)
    parser.add_argument("--api-key", type=str, help="Specify the API key for the wrapper", required=False)
    parser.add_argument("--url", type=str, help="Specify the url for the wrapper", required=False)
    parser.add_argument(
        "--model-name", type=str, help="Specify which model to run against (OpenAI and Anthropic)", required=False
    )
    parser.add_argument(
        "--az-api-version", type=str, help="Specify the Azure OpenAI API version (Azure only)", required=False
    )
    parser.add_argument("--prompt", type=str, help="Specify the prompt to use", required=False)
    parser.add_argument("--system-prompt", type=str, help="Text file containing system prompt to use.", required=False)
    parser.add_argument(
        "--selector",
        type=str,
        help="The selector to retrieve the text response from the LLM response JSON.",
        required=False,
    )
    parser.add_argument("--request-template", type=str, help="The template to wrap the API request in.", required=False)
    parser.add_argument(
        "--tokenizer",
        type=str,
        help="Choose a HuggingFace model to provide a tokeniser for prompt and chat completion templating.",
        required=False,
    )
    parser.add_argument(
        "--risk-threshold",
        type=int,
        help="Set a flagged event to total event ratio threshold above which the system will exit 1",
        required=False,
    )
    parser.add_argument(
        "--project-id",
        type=str,
        help="The project to associate with.",
        required=False,
    )



# Standard library imports
import json
import logging
import os
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import wraps
from typing import Any, Dict, List, Optional, Tuple, Union, cast

# Third party imports
import httpx
import jsonpath_ng
import requests
from anthropic import Anthropic
from anthropic.types import MessageParam
from jinja2 import Template
from openai import APIStatusError, AzureOpenAI, OpenAI, OpenAIError

# Project imports
from mindgard.exceptions import (
    EmptyResponse,
    NotImplemented,
    Unauthorized,
    Uncontactable,
    UnprocessableEntity,
    status_code_to_exception,
)
from mindgard.responses import extract_replies
from mindgard.test import RequestHandler
from mindgard.throttle import throttle
from mindgard.types import type_model_presets
from mindgard.utils import check_expected_args
from mindgard.wrappers.templating import Turn, render_messages_template

logger = logging.getLogger(__name__)


@dataclass
class PromptResponse:
    prompt: str
    response: str
    duration_ms: float


class Context:
    def __init__(self) -> None:
        self.turns: list[PromptResponse] = []

    def add(self, prompt_response: PromptResponse) -> None:
        if os.getenv("MG_DISABLE_HISTORY", "false").lower() == "true":
            # don't append
            logger.warning("⚠️ History saving is disabled, did not append to chat history!")
            return None
        self.turns.append(prompt_response)


class ContextManager:
    def __init__(self) -> None:
        self.contexts: Dict[str, Context] = {}

    def get_context_or_none(self, context_id: Optional[str] = None) -> Optional[Context]:
        if context_id is None:
            return None
        if self.contexts.get(context_id, None) is None:
            self.contexts[context_id] = Context()
        return self.contexts[context_id]


def log_time(func):
    @wraps(func)
    def _inner(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        duration = end_time - start_time
        logger.debug(f"App/Model API call took {duration * 1000.0} ms")
        return result

    return _inner


class LLMModelWrapper(ABC):
    def to_handler(self) -> RequestHandler:
        context_manager = ContextManager()

        def handler(payload: Any) -> Any:
            context = context_manager.get_context_or_none(payload.get("context_id"))
            prompt_response = self(payload["prompt"], context)
            return {"response": prompt_response.response, "duration_ms": prompt_response.duration_ms}

        return handler

    @abstractmethod
    def __call__(self, content: str, with_context: Optional[Context] = None) -> PromptResponse:
        pass


class TestStaticResponder(LLMModelWrapper):
    """
    This is only for testing
    """

    def __init__(self, system_prompt: str, handler: Any = None, rate_limit: int = 3600):
        self.context_manager = ContextManager()
        self._system_prompt = system_prompt
        self._handler = handler
        self._throttled_call = throttle(self._call_inner, rate_limit=rate_limit)

    def to_handler(self):
        if self._handler is not None:
            return self._handler
        else:
            return super().to_handler()

    @log_time
    def __call__(self, content: str, with_context: Optional[Context] = None) -> PromptResponse:
        return self._throttled_call(content, with_context)

    def _call_inner(self, content: str, with_context: Optional[Context] = None) -> PromptResponse:
        request = f"[start]sys: {self._system_prompt};"
        if with_context is not None:
            request = f"{request}"
            for prompt_response in with_context.turns:
                request = f"{request} user: {prompt_response.prompt}; assistant: {prompt_response.response};"
            request = f"{request}"
        else:
            request = f"{request}"
        request = f"{request} next: {content}[end]"

        start_time = time.time()
        response = f"TEST. prompted with: {request=}"
        duration_ms = (time.time() - start_time) * 1000

        prompt_response = PromptResponse(prompt=content, response=response, duration_ms=duration_ms)

        if with_context is not None:
            with_context.add(prompt_response)

        return prompt_response


class CustomWrapper(LLMModelWrapper):
    def __init__(
        self,
        api_url: str,
        request_template: str,
        selector: str,
        system_prompt: Optional[str] = None,
        allow_redirects: bool = True,
        headers: Optional[Dict[str, str]] = None,
        rate_limit: int = 3600,
    ) -> None:
        self._allow_redirects = allow_redirects
        self.context_manager = ContextManager()
        self.api_url = api_url
        self.throttled_call_llm = throttle(self._call_llm, rate_limit=rate_limit)
        self.template = Template(request_template)
        self.selector = selector
        self.system_prompt = system_prompt
        self.headers = headers or {}

    @log_time
    def __call__(self, content: str, with_context: Optional[Context] = None) -> str:
        return self.throttled_call_llm(content, with_context)

    def _call_llm(self, content: str, with_context: Optional[Context] = None) -> str:

        turns = []
        if with_context is not None:
            turns = [Turn(prompt=c.prompt, response=c.response) for c in with_context.turns]

        request_body = render_messages_template(
            prompt=content, template=self.template, system_prompt=self.system_prompt, turns=turns
        )

        logger.debug(f"LLM request: {request_body}")

        start_time = time.time()
        try:
            response = requests.post(
                self.api_url, headers=self.headers, data=request_body, allow_redirects=self._allow_redirects
            )
            logger.debug(f"LLM response code: {response.status_code}")
            try:
                logger.debug(f"LLM response body: {response.json()}")
            except:
                logger.debug(f"No/malformed response body")
            response.raise_for_status()
        except requests.exceptions.ConnectionError as cerr:
            raise Uncontactable(str(cerr))
        except requests.exceptions.HTTPError as httperr:
            status_code: int = httperr.response.status_code
            raise status_code_to_exception(status_code, actual_error=httperr)
        except Exception as e:
            raise e
        finally:
            duration = (time.time() - start_time) * 1000

        if response.status_code == 308:
            raise Uncontactable("Failed to contact model: model returned a 308 redirect that couldn't be followed.")

        extracted_response = extract_replies(response, self.selector)

        pr = PromptResponse(prompt=content, response=extracted_response, duration_ms=duration)

        if with_context is not None:
            with_context.add(pr)

        return pr


class APIModelWrapper(LLMModelWrapper):
    def __init__(
        self,
        api_url: str,
        allow_redirects: bool = True,
        request_template: Optional[str] = None,
        selector: Optional[str] = None,
        system_prompt: Optional[str] = None,
        tokenizer: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
        rate_limit: int = 3600,
        multi_turn_enabled: bool = False,
    ) -> None:
        self._allow_redirects = allow_redirects
        self.context_manager = ContextManager()
        self.api_url = api_url
        self.throttled_call_llm = throttle(self._call_llm, rate_limit=rate_limit)
        self.multi_turn_enabled = multi_turn_enabled
        if tokenizer:
            logging.debug("Note that with tokenizer enabled, the request_template format is different.")
            self.request_template = request_template or '{"prompt": "{tokenized_chat_template}"}'
            if "{tokenized_chat_template}" not in self.request_template:
                raise ValueError(
                    "`--request-template` must contain '{tokenized_chat_template}' when using a tokenizer."
                )
            # Third party imports
            from transformers import AutoTokenizer

            self.tokenizer = AutoTokenizer.from_pretrained(tokenizer)
        else:
            self.tokenizer = None
            default_template = (
                '{"prompt": "{system_prompt}{prompt}"}'
                if system_prompt is None
                else '{"prompt": "{system_prompt} {prompt}"}'
            )
            self.request_template = request_template or default_template
            if "{prompt}" not in self.request_template or "{system_prompt}" not in self.request_template:
                raise ValueError("`--request-template` must contain '{prompt}' and '{system_prompt}'.")
        self.selector = selector
        self.system_prompt = system_prompt or ""
        self.headers = headers or {}

    def prompt_to_request_payload(self, prompt: str) -> Dict[str, Any]:
        if self.tokenizer:
            tokenized_chat_template = cast(
                str,
                self.tokenizer.apply_chat_template(
                    [
                        {"role": "system", "content": self.system_prompt},
                        {"role": "user", "content": prompt},
                    ],
                    tokenize=False,
                ),
            )

            tokenized_chat_template = json.dumps(tokenized_chat_template, ensure_ascii=False)
            payload = self.request_template.replace("{tokenized_chat_template}", tokenized_chat_template)
        else:
            # Dump to escape quote marks that are inside the prompt/system_prompt
            prompt = json.dumps(prompt, ensure_ascii=False)
            system_prompt = json.dumps(self.system_prompt, ensure_ascii=False)
            # The dumps added extra quote marks to prompt and system prompt so trim them before the replace
            payload = self.request_template.replace("{prompt}", prompt[1:-1])
            payload = payload.replace("{system_prompt}", system_prompt[1:-1])

        logging.debug(f"{payload=}")
        # should handle non-json payload (or single string)
        payload = json.loads(payload)
        assert isinstance(
            payload, dict
        ), f"Expected the request template to form a json dict, got {type(payload)} instead."
        return cast(Dict[str, Any], payload)

    @log_time
    def __call__(self, content: str, with_context: Optional[Context] = None) -> PromptResponse:
        return self.throttled_call_llm(content, with_context)

    def _call_llm(self, content: str, with_context: Optional[Context] = None) -> PromptResponse:
        request_payload = self.prompt_to_request_payload(content)

        # Make the API call
        start_time = time.time()
        try:
            response = requests.post(
                self.api_url, headers=self.headers, json=request_payload, allow_redirects=self._allow_redirects
            )
            response.raise_for_status()
        except requests.exceptions.ConnectionError as cerr:
            raise Uncontactable(str(cerr))
        except requests.exceptions.HTTPError as httperr:
            status_code: int = httperr.response.status_code
            raise status_code_to_exception(status_code, actual_error=httperr)
        except Exception as e:
            # everything else
            raise e
        finally:
            duration = (time.time() - start_time) * 1000

        if response.status_code == 308:
            raise Uncontactable("Failed to contact model: model returned a 308 redirect that couldn't be followed.")

        extracted_response = extract_replies(response, self.selector)
        return PromptResponse(prompt=str(request_payload), response=extracted_response, duration_ms=duration)


class AzureAIStudioWrapper(APIModelWrapper):
    def __init__(
        self,
        api_key: str,
        url: str,
        request_template: Optional[str],
        system_prompt: str,
        allow_redirects: bool,
        rate_limit: Optional[int] = 60000,
    ) -> None:
        az_request_template = (
            request_template
            or """{
    "messages": [
        {"role": "system", "content": "{system_prompt}"},
        {"role": "user", "content": "{prompt}"}
    ],
    "temperature": 0.2,
    "max_tokens": 1024,
    "stop": [],
    "top_p": 0.9
}"""
        )
        self._allow_redirects = allow_redirects
        self.throttled_call_llm = throttle(self._call_llm, rate_limit=rate_limit)
        super().__init__(
            url,
            request_template=az_request_template,
            selector='["choices"][0]["message"]["content"]',
            headers={
                "Authorization": f"Bearer {api_key}",
                "Content-Type": "application/json",
            },
            system_prompt=system_prompt,
            # allow_redirects doesn't actually make any difference here, but added defensively until we can refactor away from this inheritance chain
            allow_redirects=allow_redirects,
            rate_limit=rate_limit,
        )

    @log_time
    def __call__(self, content: str, with_context: Optional[Context] = None) -> str:
        return self.throttled_call_llm(content, with_context)

    def _call_llm(self, content: str, with_context: Optional[Context] = None) -> str:
        if with_context is not None:
            logging.debug(
                "APIModelWrapper is temporarily incompatible with chat completions history. Attacks that require chat completions history fail."
            )
            raise NotImplementedError("The crescendo attack is currently incompatible with custom model wrappers.")

        request_payload = self.prompt_to_request_payload(content)

        # Make the API call
        response = requests.post(
            self.api_url, headers=self.headers, json=request_payload, allow_redirects=self._allow_redirects
        )
        if response.status_code == 400:
            try:
                # Detect Azure Content Filter error message
                err_res_json = response.json()
                if err_message := err_res_json.get("error", {}).get("message", None):
                    return cast(str, err_message)
            except Exception:
                raise status_code_to_exception(400)
        elif response.status_code == 308:
            raise Uncontactable("Failed to contact model: model returned a 308 redirect that couldn't be followed.")
        elif response.status_code != 200:
            # Handle other types of API error
            message = f"API call failed with {response.status_code=} {response.json()=}."
            raise status_code_to_exception(response.status_code)

        res_json: Dict[str, Any] = cast(Dict[str, Any], response.json())

        # Detect Cohere content_filter error
        if res_json["choices"][0]["finish_reason"] == "content_filter":
            return "Sorry, but a content filter was triggered. Please try again with a different prompt."

        # Now we are sure that the response was successful and not a content filter error
        if self.selector:
            jsonpath_expr = jsonpath_ng.parse(self.selector)
            match = jsonpath_expr.find(res_json)
            if match:
                return str(match[0].value)
            else:
                raise Exception(f"Selector {self.selector} did not match any elements in the response. {res_json=}")

        # disabled until we can support templating chat completions
        # if with_context is not None:
        #     with_context.add(PromptResponse(
        #         prompt=content,
        #         response=response
        #     ))

        return str(response)


class HuggingFaceWrapper(APIModelWrapper):
    def __init__(
        self,
        api_key: str,
        api_url: str,
        request_template: str,
        allow_redirects: bool,
        system_prompt: Optional[str] = None,
        rate_limit: Optional[int] = 60000,
        multi_turn_enabled: bool = False,
    ) -> None:
        super().__init__(
            api_url,
            request_template=request_template,
            selector='[0]["generated_text"]',
            headers={"Authorization": f"Bearer {api_key}"},
            system_prompt=system_prompt,
            rate_limit=rate_limit,
            allow_redirects=allow_redirects,
            multi_turn_enabled=multi_turn_enabled,
        )


# https://learn.microsoft.com/en-us/azure/ai-services/openai/reference#chat-completions to see available values for az_api_version
# ["2023-05-15", "2023-06-01-preview", "2023-10-01-preview", "2024-02-15-preview", "2024-03-01-preview", "2024-04-01-preview", "2024-02-01"] on May 15th 2024
class AzureOpenAIWrapper(LLMModelWrapper):
    def __init__(
        self,
        api_key: str,
        model_name: str,
        az_api_version: str,
        url: str,
        allow_redirects: bool,
        system_prompt: Optional[str] = None,
        rate_limit: Optional[int] = 60000,
    ) -> None:
        self.api_key = api_key
        self.model_name = model_name
        self.client = AzureOpenAI(
            api_key=api_key,
            api_version=az_api_version,
            azure_endpoint=url,
            http_client=httpx.Client(follow_redirects=allow_redirects),
        )
        self.system_prompt = system_prompt
        self.throttled_call_llm = throttle(self._call_llm, rate_limit=rate_limit)

    @log_time
    def __call__(self, content: str, with_context: Optional[Context] = None) -> PromptResponse:
        return self.throttled_call_llm(content, with_context)

    def _call_llm(self, content: str, with_context: Optional[Context] = None) -> PromptResponse:
        return openai_call(wrapper=self, content=content, with_context=with_context)


class OpenAIWrapper(LLMModelWrapper):
    def __init__(
        self,
        api_key: str,
        model_name: Optional[str],
        allow_redirects: bool = True,
        system_prompt: Optional[str] = None,
        api_url: Optional[str] = None,
        rate_limit: Optional[int] = 60000,
    ) -> None:
        self.api_key = api_key
        self.client = (
            OpenAI(api_key=api_key)
            if api_url is None
            else OpenAI(api_key=api_key, base_url=api_url, http_client=httpx.Client(follow_redirects=allow_redirects))
        )
        self.model_name = model_name or "gpt-3.5-turbo"
        self.system_prompt = system_prompt
        self.throttled_call_llm = throttle(self._call_llm, rate_limit=rate_limit)

    @log_time
    def __call__(self, content: str, with_context: Optional[Context] = None) -> PromptResponse:
        return self.throttled_call_llm(content, with_context)

    def _call_llm(self, content: str, with_context: Optional[Context] = None) -> PromptResponse:
        return openai_call(wrapper=self, content=content, with_context=with_context)


def openai_call(
    wrapper: Union[AzureOpenAIWrapper, OpenAIWrapper],
    content: str,
    with_context: Optional[Context] = None,
) -> PromptResponse:
    if wrapper.system_prompt:
        messages = [{"role": "system", "content": wrapper.system_prompt}]
    else:
        messages = []

    if with_context:
        for prompt_response in with_context.turns:
            messages.append({"role": "user", "content": prompt_response.prompt})
            messages.append({"role": "assistant", "content": prompt_response.response})

    messages.append({"role": "user", "content": content})

    logging.debug(messages)

    start_time = time.time()
    try:
        chat = wrapper.client.chat.completions.create(model=wrapper.model_name, messages=messages)  # type: ignore # TODO: fix type error
        response = chat.choices[0].message.content
        if not response:
            raise EmptyResponse("Model returned an empty response.")
    except APIStatusError as exception:
        if exception.status_code == 308:
            raise Uncontactable(
                "Failed to contact model: model returned a 308 redirect that couldn't be followed."
            ) from exception
        if exception.status_code == 401:
            raise Unauthorized(
                "Failed to contact model: model returned a 401 (unauthorized).", status_code=401
            ) from exception
        if exception.status_code == 422:
            err_message = "<none>"
            try:
                err_message = exception.response.json().get("error", err_message)
            except Exception:
                pass
            raise UnprocessableEntity(f"Received 422 from provider: {err_message}", 422) from exception
        raise status_code_to_exception(exception.status_code) from exception
    except OpenAIError as exception:
        raise EmptyResponse("An OpenAI error occurred") from exception
    finally:
        duration_ms = (time.time() - start_time) * 1000

    prompt_response = PromptResponse(prompt=content, response=response, duration_ms=duration_ms)

    if with_context is not None:
        with_context.add(prompt_response)

    return prompt_response


class AnthropicWrapper(LLMModelWrapper):
    def __init__(
        self,
        api_key: str,
        model_name: Optional[str],
        system_prompt: Optional[str] = None,
        rate_limit: Optional[int] = 60000,
    ) -> None:
        self.api_key = api_key
        self.client = Anthropic(api_key=api_key)
        self.model_name = model_name or "claude-3-opus-20240229"
        self.system_prompt = system_prompt
        self.throttled_call_llm = throttle(self._call_llm, rate_limit=rate_limit)

    @log_time
    def __call__(self, content: str, with_context: Optional[Context] = None) -> PromptResponse:
        return self.throttled_call_llm(content, with_context)

    def _call_llm(self, content: str, with_context: Optional[Context] = None) -> PromptResponse:
        messages: List[MessageParam] = []

        if with_context:
            for prompt_response in with_context.turns:
                messages.append({"role": "user", "content": prompt_response.prompt})
                messages.append({"role": "assistant", "content": prompt_response.response})

        messages.append({"role": "user", "content": content})

        start_time = time.time()
        if self.system_prompt is not None:
            message = self.client.messages.create(
                system=self.system_prompt,
                max_tokens=1024,
                messages=messages,
                model=self.model_name,
            )
        else:
            message = self.client.messages.create(max_tokens=1024, messages=messages, model=self.model_name)
        response = message.content[0].text

        duration_ms = (time.time() - start_time) * 1000

        prompt_response = PromptResponse(prompt=content, response=response, duration_ms=duration_ms)

        if with_context is not None:
            with_context.add(prompt_response)

        return prompt_response


def get_llm_model_wrapper(
    headers: Dict[str, str],
    preset: Optional[type_model_presets] = None,
    api_key: Optional[str] = None,
    url: Optional[str] = None,
    model_name: Optional[str] = None,
    az_api_version: Optional[str] = None,
    system_prompt: Optional[str] = None,
    selector: Optional[str] = None,
    request_template: Optional[str] = None,
    tokenizer: Optional[str] = None,
    rate_limit: int = 60000,
    allow_redirects: bool = True,
    force_multi_turn: bool = False,
) -> LLMModelWrapper:
    # Create model based on preset
    if preset == "huggingface-openai" or preset == "openai-compatible":
        check_expected_args(locals(), ["api_key", "url"])
        api_key, url, request_template = cast(Tuple[str, str, str], (api_key, url, request_template))
        if url[-4:] != "/v1/":
            url += "/v1/"
        return OpenAIWrapper(
            api_key=api_key,
            api_url=url,
            system_prompt=system_prompt,
            model_name=model_name or "tgi",
            allow_redirects=allow_redirects,
            rate_limit=rate_limit,
        )
    elif preset == "huggingface":
        check_expected_args(locals(), ["api_key", "url", "request_template"])
        api_key, url, request_template = cast(Tuple[str, str, str], (api_key, url, request_template))
        return HuggingFaceWrapper(
            api_key=api_key,
            api_url=url,
            system_prompt=system_prompt,
            request_template=request_template,
            rate_limit=rate_limit,
            allow_redirects=allow_redirects,
            multi_turn_enabled=force_multi_turn,
        )
    elif preset == "azure-aistudio":
        check_expected_args(locals(), ["api_key", "url", "system_prompt"])
        api_key, url, system_prompt = cast(Tuple[str, str, str], (api_key, url, system_prompt))
        return AzureAIStudioWrapper(
            api_key=api_key,
            url=url,
            request_template=request_template,
            system_prompt=system_prompt,
            allow_redirects=allow_redirects,
            rate_limit=rate_limit,
        )
    elif preset == "openai":
        check_expected_args(locals(), ["api_key"])
        api_key = cast(str, api_key)
        return OpenAIWrapper(api_key=api_key, model_name=model_name, system_prompt=system_prompt, rate_limit=rate_limit)
    elif preset == "azure-openai":
        check_expected_args(locals(), ["api_key", "model_name", "az_api_version", "url"])
        api_key, model_name, az_api_version, url = cast(
            Tuple[str, str, str, str], (api_key, model_name, az_api_version, url)
        )
        return AzureOpenAIWrapper(
            api_key=api_key,
            model_name=model_name,
            az_api_version=az_api_version,
            url=url,
            system_prompt=system_prompt,
            allow_redirects=allow_redirects,
            rate_limit=rate_limit,
        )
    elif preset == "anthropic":
        if not api_key:
            raise ValueError("`--api-key` argument is required when using the 'anthropic' preset.")
        return AnthropicWrapper(
            api_key=api_key, model_name=model_name, system_prompt=system_prompt, rate_limit=rate_limit
        )
    elif preset == "tester":
        if not system_prompt:
            raise ValueError("`--system-prompt` argument is required")
        return TestStaticResponder(system_prompt=system_prompt, rate_limit=rate_limit)

    elif preset == "custom":
        return CustomWrapper(
            api_url=url,
            request_template=request_template,
            selector=selector,
            system_prompt=system_prompt,
            allow_redirects=allow_redirects,
            headers=headers,
            rate_limit=rate_limit,
        )

    else:
        if not url:
            raise ValueError("`--url` argument is required when not using a preset configuration.")
        # Convert headers string to dictionary
        if headers:
            return APIModelWrapper(
                api_url=url,
                allow_redirects=allow_redirects,
                selector=selector,
                request_template=request_template,
                system_prompt=system_prompt,
                tokenizer=tokenizer,
                headers=headers,
                rate_limit=rate_limit,
                multi_turn_enabled=force_multi_turn,
            )
        else:
            return APIModelWrapper(
                api_url=url,
                allow_redirects=allow_redirects,
                selector=selector,
                request_template=request_template,
                system_prompt=system_prompt,
                tokenizer=tokenizer,
                rate_limit=rate_limit,
                multi_turn_enabled=force_multi_turn,
            )



# Standard library imports
import csv
import json
import os
import sys
from argparse import Namespace
from typing import Any, Dict, List, Optional, Tuple

# Third party imports
import requests
import toml
from rich.console import Console

# Project imports
from mindgard.constants import (
    REPOSITORY_URL,
    VERSION,
)
from mindgard.orchestrator import GetTestAttacksResponse, OrchestratorTestResponse
from mindgard.types import type_model_presets_list, valid_llm_datasets


class CliResponse:
    def __init__(self, code: int):
        self._code = code

    def code(self) -> int:
        return self._code


class ConflictError(ValueError):
    def __init__(self, message: str):
        super().__init__(message)


def convert_test_to_cli_response(test: GetTestAttacksResponse, risk_threshold: int) -> CliResponse:
    try:
        flagged_to_total_events_ratio = test.test.flagged_events / test.test.total_events
    except ZeroDivisionError:
        flagged_to_total_events_ratio = 0.0
    if flagged_to_total_events_ratio >= risk_threshold:
        return CliResponse(1)
    else:
        return CliResponse(0)


def print_to_stderr(*args: Any, **kwargs: Any) -> None:
    print(*args, file=sys.stderr, **kwargs)


def print_to_stderr_as_json(dict: Any) -> None:
    print(json.dumps(dict), file=sys.stderr)


def version_to_tuple(version: str) -> Tuple[int, ...]:
    return tuple(map(int, version.split(".")))


def is_version_outdated() -> Optional[str]:
    try:
        res = requests.get(REPOSITORY_URL)
        res.raise_for_status()
        latest_version = res.json()["info"]["version"]
        latest_version_tuple = version_to_tuple(latest_version)
        current_version_tuple = version_to_tuple(VERSION)
        return latest_version if latest_version_tuple > current_version_tuple else None
    except Exception:
        return ""


def map_mode_to_attack_pack(mode: str) -> str:
    if mode == "exhaustive":
        return "large"
    elif mode == "thorough":
        return "medium"
    else:
        return "sandbox"


def parse_toml_and_args_into_final_args(config_file_path: Optional[str], args: Namespace) -> Dict[str, Any]:
    config_file = config_file_path or "mindgard.toml"
    toml_args = {}
    try:
        with open(config_file, "r") as f:
            contents = f.read()
            toml_args = toml.loads(contents)
    except FileNotFoundError:
        if config_file_path is None:
            pass
        else:
            raise ValueError(f"Config file not found: {config_file=}. Check that the file exists on disk.")

    final_args: Dict[str, Any] = dict(toml_args)

    for k, v in vars(args).items():
        if v is not None and k in final_args:
            formatted_key = k.replace("_", " ")
            if k == "project_id":
                formatted_key = "Project ID"
            raise ConflictError(f"Error: Conflicting {formatted_key}")
        elif v is not None:
            final_args[k] = v
        else:
            final_args[k] = toml_args.get(k) or toml_args.get(k.replace("_", "-"))

    preset = final_args["preset"]
    if preset is not None and preset not in type_model_presets_list:
        raise ValueError(f"Preset '{preset}' is invalid, must be one of {type_model_presets_list}")

    final_args["api_key"] = final_args["api_key"] or os.environ.get("MODEL_API_KEY", None)

    final_args["risk_threshold"] = (
        final_args.get("risk_threshold") if final_args.get("risk_threshold") is not None else 50
    )
    final_args["parallelism"] = (
        final_args.get("parallelism")
        if final_args.get("parallelism") is not None
        else final_args.pop("parallelism", None)
    )
    final_args["json"] = final_args.get("json") if final_args.get("json") is not None else False
    final_args["mode"] = final_args.get("mode") if final_args.get("mode") is not None else "fast"
    final_args["rate_limit"] = (
        final_args.get("rate_limit") if final_args.get("rate_limit") is not None else 3600
    )  # 60rps

    if args.command == "test":
        final_args["attack_pack"] = map_mode_to_attack_pack(args.mode)

    domain = final_args.get("domain", None)
    dataset = final_args.get("dataset", None)

    if dataset is None:
        if domain and (domain not in valid_llm_datasets.keys()):
            raise ValueError(
                f"Domain set in config file ({final_args['domain']}) was invalid! (choices: {[x for x in valid_llm_datasets]})"
            )
        final_args["dataset"] = valid_llm_datasets.get(domain, None)
    else:

        if not os.path.exists(dataset):
            raise ValueError(
                f"Dataset {dataset} not found! Please provide a valid path to a dataset with new line separated prompts."
            )

        with open(dataset, "r") as datasets_file:
            try:
                datasets = csv.reader(datasets_file)
                all_rows = [line for line in datasets]

                non_empty_rows = list(filter(lambda row: row != [], all_rows))

                if len(non_empty_rows) == 0:
                    raise ValueError("Custom dataset should not be an empty file!")

                # Enforce a one-column CSV and raise an error if two columns are provided, as this indicates malformed datasets.
                lines = []
                for i, line in enumerate(non_empty_rows):
                    if len(line) > 1:
                        raise ValueError(f"Error on CSV row {i+1}: custom dataset only supports one column per line")
                    lines.append(line[0].strip())
            except csv.Error:
                raise ValueError(f"{dataset} is not a valid CSV file!")

            final_args["custom_dataset"] = json.dumps(lines)

    if final_args.get("force_multi_turn"):
        final_args["parallelism"] = 1

    return final_args


def check_expected_args(args: Dict[str, Any], expected_args: List[str]) -> None:
    missing_args: List[str] = []
    for arg in expected_args:
        if not args.get(arg):
            missing_args.append(f"`--{arg.replace('_', '-')}`")
    if missing_args:
        raise ValueError(f"Missing required arguments: {', '.join(missing_args)}")


def request_user_input(console: Console, message: str, valid_responses: List[str]) -> str:
    """
    Request console input from the user and returns it if it's valid.

    Args:
        console (Console): Console object to use for input.
        message (str): Request message to display to the user.
        valid_responses (List[str]): Valid responses to accept.

    Returns:
        str: Accepted response from the user.
    """
    while True:
        response = console.input(message).strip().lower()
        if response in valid_responses:
            return response
        else:
            print(f"Invalid response. Please enter one of: {', '.join(valid_responses)}")


def invalid_project_id_error(console, error: str):
    if "project id" in error.lower():
        console.print(
            """\nRun [blue]`mindgard list projects`[/blue] to view valid projects, or manage them in the Web UI.\n\nVisit https://docs.mindgard.ai/user-guide/projects for more details.\n"""
        )


def unspecified_project_id_error(console):
    console.print(f"{'[red bold]Error: Unspecified Project ID' }")
    console.print(
        """\nAdd a project id in your config file or pass it with the command [blue]--project <projectID>[/blue]\nRun [blue]`mindgard list projects`[/blue] to view available projects, or manage them in the Web UI.
        """
    )
    console.print("Visit https://docs.mindgard.ai/user-guide/projects for more details.\n")


def conflicting_args_error(console, error):
    console.print(f"[red bold]{error}")
    console.print("\nOnly one should be provided, either in your command line or config file.\n")



# Standard library imports
import json
import logging
import random
import re
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Optional, Tuple

# Third party imports
import requests
from rich.console import Console

# Project imports
from mindgard.wrappers.llm import (
    AnthropicWrapper,
    APIModelWrapper,
    AzureAIStudioWrapper,
    AzureOpenAIWrapper,
    Context,
    HuggingFaceWrapper,
    LLMModelWrapper,
    OpenAIWrapper,
    PromptResponse,
    TestStaticResponder,
)

logger = logging.getLogger(__name__)


class CheckStatus(Enum):
    PASS = "pass"
    FAIL = "fail"
    SKIP = "skip"
    ERROR = "error"


@dataclass
class CompatCheckResult:
    """Result of a single compatibility check."""

    name: str
    status: CheckStatus
    message: str
    details: Optional[str] = None


@dataclass
class CompatReport:
    """Full compatibility report containing all check results."""

    # Wrapper checks
    openai_static: CompatCheckResult
    openai_runtime: CompatCheckResult

    # Technique checks
    session_memory: CompatCheckResult
    session_isolation: CompatCheckResult
    session_overwriting: CompatCheckResult

    @property
    def all_passed(self) -> bool:
        """Check if all non-skipped checks passed."""
        results = [
            self.openai_static,
            self.openai_runtime,
            self.session_memory,
            self.session_isolation,
            self.session_overwriting,
        ]
        # A check counts as "passed" if it's PASS or SKIP
        return all(r.status in (CheckStatus.PASS, CheckStatus.SKIP) for r in results)


CONVERSATION_CAPABLE_WRAPPERS = (
    OpenAIWrapper,
    AzureOpenAIWrapper,
    AnthropicWrapper,
    TestStaticResponder,
)

SINGLE_TURN_WRAPPERS = (
    APIModelWrapper,
    HuggingFaceWrapper,
    AzureAIStudioWrapper,
)

OPENAI_URL_PATTERNS = [
    r".*/v1/chat/completions/?$",
    r".*/chat/completions/?$",
]


def _test_phrase_1() -> str:
    return "carrot-apple"


def _test_phrase_2() -> str:
    return "yam-strawberry"


def check_openai_static(wrapper: LLMModelWrapper) -> CompatCheckResult:
    """
    Static check: Determine if the wrapper class supports conversations.

    This checks the wrapper TYPE, not runtime behavior.
    """
    wrapper_type = type(wrapper).__name__

    if isinstance(wrapper, CONVERSATION_CAPABLE_WRAPPERS):
        return CompatCheckResult(
            name="OpenAI Static Check",
            status=CheckStatus.PASS,
            message=f"Wrapper '{wrapper_type}' supports conversations",
            details="This wrapper builds message arrays with conversation history",
        )

    if isinstance(wrapper, SINGLE_TURN_WRAPPERS):
        return CompatCheckResult(
            name="OpenAI Static Check",
            status=CheckStatus.FAIL,
            message=f"Wrapper '{wrapper_type}' does not support conversations",
            details="This wrapper sends prompts without conversation history",
        )

    # Unknown wrapper type
    return CompatCheckResult(
        name="OpenAI Static Check",
        status=CheckStatus.FAIL,
        message=f"Unknown wrapper type '{wrapper_type}' - cannot determine conversation support",
        details="Consider using a known preset (openai, anthropic, azure-openai)",
    )


def _url_matches_openai_pattern(url: Optional[str]) -> bool:
    """Check if URL matches OpenAI-style chat completion endpoint pattern."""
    if not url:
        return False
    for pattern in OPENAI_URL_PATTERNS:
        if re.match(pattern, url, re.IGNORECASE):
            return True
    return False


def _try_openai_request(url: str, headers: Optional[Dict[str, str]] = None) -> Tuple[bool, Optional[str]]:
    """
    Try sending an OpenAI-style chat completion request to the URL.

    Returns (is_openai_compatible, response_content_or_error)
    """
    payload = {"messages": [{"role": "user", "content": "hello!"}]}

    try:
        response = requests.post(url, json=payload, headers=headers or {}, timeout=30)
        response.raise_for_status()
        data = response.json()

        # Check for OpenAI-style response structure
        if "choices" in data and isinstance(data["choices"], list) and len(data["choices"]) > 0:
            choice = data["choices"][0]
            if "message" in choice and "content" in choice.get("message", {}):
                content = choice["message"]["content"]
                return True, content

        return False, "Response does not match OpenAI chat completion format"
    except requests.exceptions.Timeout:
        return False, "Request timed out"
    except requests.exceptions.ConnectionError as e:
        return False, f"Connection error: {str(e)}"
    except requests.exceptions.HTTPError as e:
        return False, f"HTTP error: {e.response.status_code}"
    except Exception as e:
        return False, f"Error: {str(e)}"


def check_openai_runtime(
    wrapper: LLMModelWrapper,
    url: Optional[str],
    headers: Optional[Dict[str, str]] = None,
) -> CompatCheckResult:
    """
    Check if a URL that looks like OpenAI endpoint is actually OpenAI-compatible.
    """
    # Skip if already using an OpenAI-compatible wrapper
    if isinstance(wrapper, (OpenAIWrapper, AzureOpenAIWrapper)):
        return CompatCheckResult(
            name="OpenAI Runtime Check",
            status=CheckStatus.SKIP,
            message="Already using OpenAI-compatible wrapper",
            details=None,
        )

    # Check if URL matches OpenAI pattern
    if url is None or not _url_matches_openai_pattern(url):
        return CompatCheckResult(
            name="OpenAI Runtime Check",
            status=CheckStatus.SKIP,
            message="URL does not match OpenAI pattern",
            details="URL pattern check only applies to endpoints like */v1/chat/completions",
        )

    # URL matches pattern - try an actual request
    is_compatible, result = _try_openai_request(url, headers)

    if is_compatible:
        fixed_url = url.replace("/v1/chat/completions", "")
        return CompatCheckResult(
            name="OpenAI Runtime Check",
            status=CheckStatus.PASS,
            message="URL appears to be OpenAI-compatible!",
            details=f"Recommendation: Use --preset openai-compatible and '{fixed_url}' url instead for better conversation support",
        )
    else:
        return CompatCheckResult(
            name="OpenAI Runtime Check",
            status=CheckStatus.FAIL,
            message="URL matches OpenAI pattern but endpoint is not compatible",
            details=result,
        )


def check_session_is_appendable(
    wrapper: LLMModelWrapper,
) -> CompatCheckResult:
    """
    Session memory check: Send a 2-turn conversation and verify the second turn
    receives the context from the first turn.

    Uses an "echo test" pattern:
    Turn 1: Ask the model to remember a unique phrase
    Turn 2: Ask the model to recall the phrase
    """
    test_phrase = _test_phrase_1()

    turn1_prompt = f"I'm going to tell you a made-up phrase. Please remember it exactly. The phrase is: {test_phrase}. Just acknowledge you received it."
    turn2_prompt = "What was the phrase I just told you? Please respond with ONLY the phrase, nothing else."

    context = Context()

    try:
        logger.debug(f"Session memory check - Turn 1: {turn1_prompt}")
        response1 = wrapper(turn1_prompt, with_context=context)
        logger.debug(f"Turn 1 response: {response1.response}")

        logger.debug(f"Session memory check - Turn 2: {turn2_prompt}")
        response2 = wrapper(turn2_prompt, with_context=context)
        logger.debug(f"Turn 2 response: {response2.response}")

        if test_phrase in response2.response.lower():
            return CompatCheckResult(
                name="Session Memory",
                status=CheckStatus.PASS,
                message="Model correctly recalled information from previous turn",
                details=f"Test phrase '{test_phrase}' was successfully echoed back",
            )
        else:
            return CompatCheckResult(
                name="Session Memory",
                status=CheckStatus.FAIL,
                message="Model did NOT recall information from previous turn",
                details=f"Expected '{test_phrase}' in response, got: '{response2.response[:100]}...'",
            )

    except NotImplementedError as e:
        return CompatCheckResult(
            name="Session Memory",
            status=CheckStatus.FAIL,
            message="Wrapper raised NotImplementedError for conversations",
            details=str(e),
        )
    except Exception as e:
        logger.error(f"Session memory check failed with exception: {e}")
        return CompatCheckResult(
            name="Session Memory",
            status=CheckStatus.ERROR,
            message=f"Error during session memory check: {type(e).__name__}",
            details=str(e),
        )


def check_session_is_reset(
    wrapper: LLMModelWrapper,
) -> CompatCheckResult:
    """
    Session isolation check: Test if the target model has persistent memory across
    SEPARATE conversation contexts.

    This checks if the model retains information from a PREVIOUS context
    when asked in a NEW context (i.e., no explicit context passed).

    Most models should NOT have this - it would indicate shared state.
    """
    test_phrase = _test_phrase_1()

    setup_prompt = f"Remember this phrase for later: {test_phrase}. Just say 'Noted' to confirm."
    recall_prompt = "Do you remember any fruit-vegetable phrase I mentioned recently? If yes, what is it? If no, say 'No memory of that'."

    try:
        # Context 1: Store the code word
        context1 = Context()
        logger.debug(f"Session isolation check - Setup in context1: {setup_prompt}")
        response1 = wrapper(setup_prompt, with_context=context1)
        logger.debug(f"Setup response: {response1.response}")

        # Context 2 (NEW, separate context): Try to recall
        context2 = Context()
        logger.debug(f"Session isolation check - Recall in new context2: {recall_prompt}")
        response2 = wrapper(recall_prompt, with_context=context2)
        logger.debug(f"Recall response: {response2.response}")

        # If the phrase appears, model has unexpected persistent memory
        if test_phrase in response2.response.lower():
            return CompatCheckResult(
                name="Session Isolation",
                status=CheckStatus.FAIL,
                message="WARNING: Model has persistent memory across sessions!",
                details="The model remembered information from a previous session - this may indicate shared state",
            )
        else:
            return CompatCheckResult(
                name="Session Isolation",
                status=CheckStatus.PASS,
                message="Model does NOT appear to have persistent memory across sessions",
                details="Fresh sessions start without previous conversation history",
            )

    except Exception as e:
        logger.error(f"Session isolation check failed with exception: {e}")
        return CompatCheckResult(
            name="Session Isolation",
            status=CheckStatus.ERROR,
            message=f"Error during session isolation check: {type(e).__name__}",
            details=str(e),
        )


def check_session_is_overwritable(
    wrapper: LLMModelWrapper,
) -> CompatCheckResult:
    """
    Assert that the CLI can send the target a user-assistant message history and have the target abide by this.
    Implies full control over the targets history.

    Dummy Turn 1: The user sends a message
    Dummy Turn 2: The target response is mocked with an embedded phrase
    Dummy Turn 3: The user sends another message asking the target repeat its previous message
    Turn 4: The target responds with something containing the embedded phrase

    <same session>

    Dummy Turn 1: The user sends a message
    Dummy Turn 2: The target response is mocked with an embedded phrase
    Dummy Turn 3: The user sends another message asking the target repeat its previous message
    Turn 4: The target responds with something containing the embedded phrase
    """

    initial_user_prompt = "What's the capital of France?"
    assistant_response_template = "The capital of France is {test_phrase}."
    follow_up_user_prompt = "What was your last message to me?"

    pre_overwrite_test_phrase = _test_phrase_1()
    post_overwrite_test_phrase = _test_phrase_2()

    context = Context()
    context.add(
        PromptResponse(
            prompt=initial_user_prompt,
            response=assistant_response_template.format(test_phrase=pre_overwrite_test_phrase),
            duration_ms=0.0,
        )
    )

    try:
        assistant_r1 = wrapper(follow_up_user_prompt, with_context=context)

        if pre_overwrite_test_phrase in assistant_r1.response.lower():

            context.turns = []
            context.add(
                PromptResponse(
                    prompt=initial_user_prompt,
                    response=assistant_response_template.format(test_phrase=post_overwrite_test_phrase),
                    duration_ms=0.0,
                )
            )

            assistant_r2 = wrapper(follow_up_user_prompt, with_context=context)

            if post_overwrite_test_phrase in assistant_r2.response.lower():
                return CompatCheckResult(
                    name="History overwriting",
                    status=CheckStatus.PASS,
                    message="Model had its history overwritten",
                    details=f"Test phrase '{post_overwrite_test_phrase}' was successfully echoed back after overwriting test phrase '{pre_overwrite_test_phrase}'.",
                )
            else:
                return CompatCheckResult(
                    name="History overwriting step 2",
                    status=CheckStatus.FAIL,
                    message="Model did not have its history overwritten successfully",
                    details=f"Expected '{post_overwrite_test_phrase}' in response, got: '{assistant_r1.response[:100]}...'",
                )

        else:
            return CompatCheckResult(
                name="History overwriting step 1",
                status=CheckStatus.FAIL,
                message="Model has could not remember initial test phrase making overwriting impossible",
                details=f"Expected '{pre_overwrite_test_phrase}' in response, got: '{assistant_r1.response[:100]}...'",
            )

    except NotImplementedError as e:
        return CompatCheckResult(
            name="History overwriting",
            status=CheckStatus.FAIL,
            message="Wrapper raised NotImplementedError for conversations",
            details=str(e),
        )
    except Exception as e:
        logger.error(f"History overwriting check failed with exception: {e}")
        return CompatCheckResult(
            name="History overwriting",
            status=CheckStatus.ERROR,
            message=f"Error during session memory check: {type(e).__name__}",
            details=str(e),
        )


def run_compat_checks(
    wrapper: LLMModelWrapper,
    url: Optional[str] = None,
    headers: Optional[Dict[str, str]] = None,
    skip_runtime: bool = False,
) -> CompatReport:
    """
    Run all compatibility checks and return a full report.

    Args:
        wrapper: The LLM model wrapper to test
        console: Rich console for output
        url: The target URL (for OpenAI runtime check)
        headers: Request headers (for OpenAI runtime check)
        json_out: Whether to output in JSON format
        skip_runtime: Skip runtime tests (only do static checks)

    Returns:
        CompatReport with results of all checks
    """
    # Wrapper checks
    # 1. OpenAI static check (always runs)
    openai_static_result = check_openai_static(wrapper)

    # 2. OpenAI runtime check (runs if URL matches pattern)
    openai_runtime_result = check_openai_runtime(wrapper, url, headers)

    # Technique checks
    # 3. Session memory check
    if skip_runtime:
        session_memory_result = CompatCheckResult(
            name="Session Memory",
            status=CheckStatus.SKIP,
            message="Skipped (--skip-runtime flag)",
            details=None,
        )
    elif openai_static_result.status == CheckStatus.FAIL and openai_runtime_result.status != CheckStatus.PASS:
        # If static check failed AND OpenAI runtime didn't find a compatible endpoint,
        # skip session memory check (it will fail anyway)
        session_memory_result = CompatCheckResult(
            name="Session Memory",
            status=CheckStatus.SKIP,
            message="Skipped (static check failed)",
            details="Session memory check skipped because wrapper doesn't support conversations",
        )
    else:
        session_memory_result = check_session_is_appendable(wrapper)

    # 4. Session isolation check
    if skip_runtime:
        session_isolation_result = CompatCheckResult(
            name="Session Isolation",
            status=CheckStatus.SKIP,
            message="Skipped (--skip-runtime flag)",
            details=None,
        )
    else:
        session_isolation_result = check_session_is_reset(wrapper)

    if skip_runtime:
        session_overwritable_result = CompatCheckResult(
            name="History overwriting", status=CheckStatus.SKIP, message="Skipped (--skip-runtime flag)", details=None
        )
    else:
        session_overwritable_result = check_session_is_overwritable(wrapper)

    return CompatReport(
        openai_static=openai_static_result,
        openai_runtime=openai_runtime_result,
        session_memory=session_memory_result,
        session_isolation=session_isolation_result,
        session_overwriting=session_overwritable_result,
    )



# Standard library imports
from collections import namedtuple

Recon_commands = namedtuple(
    "Recon_commands",
    [
        "recon",
        "guardrail",
        "input_encoding",
        "output_encoding",
        "non_contextual",
        "output_formatting",
        "output_rendering",
        "code_generation",
        "system_prompt_extraction",
        "tool_discovery",
    ],
)

recon_constants = Recon_commands(
    recon="recon",
    guardrail="guardrail",
    input_encoding="input-encoding",
    output_encoding="output-encoding",
    non_contextual="non-contextual",
    output_formatting="output-formatting",
    output_rendering="output-rendering",
    code_generation="code-generation",
    system_prompt_extraction="system-prompt-extraction",
    tool_discovery="tool-discovery",
)



class ClientException(Exception):
    status_code: int
    message: str

    def __init__(self, message: str, status_code: int):
        self.status_code = status_code
        self.message = message

    def __str__(self):
        return f"[{self.status_code}] {self.message}"



# Standard library imports
from datetime import datetime
from typing import List, Optional

# Third party imports
from rich.progress import Progress
from rich.table import Table

from ..api_service import api_get
from ..constants import DASHBOARD_URL
from ..orchestrator import GetTestListResponse, get_tests
from ..types import (
    type_ui_exception_map,
    type_ui_task_map,
)
from ..utils import print_to_stderr_as_json


def list_test_submit(
    access_token: str,
    ui_exception_map: type_ui_exception_map,
    ui_exception_progress: Progress,
) -> GetTestListResponse:
    tests_res = get_tests(access_token, request_function=api_get)

    return tests_res


def list_test_polling(
    access_token: str,
    tests_response: GetTestListResponse,
    ui_task_map: type_ui_task_map,
    ui_task_progress: Progress,
) -> Optional[GetTestListResponse]:
    return tests_response


def list_test_output(test_list: GetTestListResponse, json_out: bool) -> Optional[Table]:
    if json_out:
        print_to_stderr_as_json([test.model_dump() for test in test_list.items])
        return None
    else:
        table = Table(title=f"Tests run by user")
        table.add_column("Model Name", style="cyan")
        table.add_column("Assessment Date", style="magenta")
        table.add_column("Flagged Events", justify="right", style="green")
        table.add_column("URL", justify="right", style="green")

        for test in test_list.items:
            table.add_row(
                test.project_id,
                datetime.fromisoformat(test.created_at.replace("Z", "+00:00")).strftime("%H:%M, %Y-%m-%d"),
                f"{test.flagged_events} / {test.total_events}",
                f"{DASHBOARD_URL}/r/test/{test.id}/{test.project_id}",
            )
        return table



# Standard library imports
from typing import Any, Dict

# Third party imports
import requests
from requests import HTTPError
from tenacity import retry, stop_after_attempt, wait_fixed

from .constants import (
    API_RETRY_ATTEMPTS,
    API_RETRY_WAIT_BETWEEN_ATTEMPTS_SECONDS,
    VERSION,
)


class MGHTTPError(HTTPError):
    def __init__(self, response: requests.Response):
        self.response = response
        self.status_code = response.status_code

        try:
            body = response.json()
        except ValueError:
            body = response.text

        self.body = body
        self.message = body.get("error") if isinstance(body, dict) and "error" in body else str(body)

        super().__init__(self.message, response=response)

    def __str__(self) -> str:
        return f"[{self.status_code}] {self.message}"


def _standard_headers(access_token: str) -> Dict[str, str]:
    return {
        "Authorization": f"Bearer {access_token}",
        "User-Agent": f"mindgard-cli/{VERSION}",
        "X-User-Agent": f"mindgard-cli/{VERSION}",
    }


@retry(
    stop=stop_after_attempt(API_RETRY_ATTEMPTS),
    wait=wait_fixed(API_RETRY_WAIT_BETWEEN_ATTEMPTS_SECONDS),
    reraise=True,
)
def api_post(url: str, access_token: str, payload: Dict[str, Any]) -> requests.Response:
    try:
        response = requests.post(url=url, json=payload, headers=_standard_headers(access_token))
        response.raise_for_status()
        return response
    except HTTPError as e:
        raise MGHTTPError(e.response) from e


@retry(
    stop=stop_after_attempt(API_RETRY_ATTEMPTS),
    wait=wait_fixed(API_RETRY_WAIT_BETWEEN_ATTEMPTS_SECONDS),
    reraise=True,
)
def api_get(url: str, access_token: str) -> requests.Response:
    response = requests.get(url=url, headers=_standard_headers(access_token))
    response.raise_for_status()
    return response



# Standard library imports
import logging
import uuid
from time import sleep
from typing import Optional

# Project imports
from mindgard.api.recon.recon import ReconClient, ReconResult, StartUnifiedReconRequest
from mindgard.api.recon.recon_prompt_events import (
    PopEventRequest,
    PopEventResponse,
    PromptResult,
    PushEventRequest,
)
from mindgard.api.recon.types import GetReconRequest
from mindgard.exceptions import MGException
from mindgard.wrappers.llm import LLMModelWrapper


class ReconCommand:
    def __init__(self, service: ReconClient, call_system_under_test: LLMModelWrapper, type: str):
        self.call_system_under_test = call_system_under_test
        self.service = service
        self.type = type

    def start_recon(self, project_id: str) -> uuid.UUID:
        f"Starting {self.type} recon for target: {str(project_id)}"
        response = self.service.start_recon(StartUnifiedReconRequest(type=self.type))
        return response.recon_id

    def poll_recon(self, recon_id: uuid.UUID) -> None:
        return self.poll(recon_id)

    def poll(self, recon_id: uuid.UUID) -> None:
        completed = False
        event: Optional[PopEventResponse] = None

        logging.debug(f"Polling for prompt requests to process: {str(recon_id)}")
        while not completed:
            completed, event = self._poll_inner(recon_id)
            if completed:
                break  # dont sleep again if completed
            sleep(0.5)

        logging.debug(f"All prompt requests processed for recon_id: {str(recon_id)}")
        logging.debug("Fetching reconnaissance result...")

    def _poll_inner(self, recon_id: uuid.UUID) -> tuple[bool, Optional[PopEventResponse]]:
        request = PopEventRequest(
            source_id=recon_id,
            event_type=["prompt_request", "complete"],
        )

        event = self.service.pop(request)

        if event is not None:
            if event.event_type == "complete":
                logging.debug(f"Reconnaissance completed for recon_id: {str(recon_id)}")
                return True, event

            if event.event_type == "prompt_request" and len(event.prompt_request) > 0:
                logging.debug(f"Processing prompt request: {str(recon_id)}")
                for request in event.prompt_request:
                    response = None
                    exception = None
                    try:
                        response = self.call_system_under_test(request.prompt)
                    except MGException as e:
                        logging.debug(f"Error communicating with target application: {e}")
                        exception = e

                    content = response.response if response else None
                    duration_ms = response.duration_ms if response else None
                    error_code = None if exception is None else getattr(exception, "status_code", None)
                    error_message = None if exception is None else str(exception)

                    logging.debug(f"Pushing prompt result for request")
                    push_event = PushEventRequest(
                        source_id=event.source_id,
                        event_type="prompt_result",
                        prompt_result=[
                            PromptResult(
                                id=request.id,
                                content=content,
                                duration_ms=duration_ms,
                                prompt_request=request,
                                error_code=error_code,
                                error_message=error_message,
                            )
                        ],
                    )

                    self.service.push(push_event)
                return False, event

        return False, None

    def fetch_recon_result(self, recon_id: uuid.UUID) -> ReconResult:
        return self.service.get_recon_result(GetReconRequest(recon_id=recon_id))



# Standard library imports
import os

# Project imports
from mindgard.config import load_auth_config

from .version import VERSION as v_VERSION

VERSION: str = v_VERSION

REPOSITORY_URL: str = f"https://pypi.org/pypi/mindgard/json"

ALGORITHMS = ["RS256"]

auth_config = load_auth_config()
AUTH0_DOMAIN = auth_config.AUTH0_DOMAIN
AUTH0_CLIENT_ID = auth_config.AUTH0_CLIENT_ID
AUTH0_AUDIENCE = auth_config.AUTH0_AUDIENCE
API_BASE = auth_config.API_BASE
DASHBOARD_URL = auth_config.DASHBOARD_URL

API_RETRY_ATTEMPTS = 3
API_RETRY_WAIT_BETWEEN_ATTEMPTS_SECONDS = 3
DEFAULT_RISK_THRESHOLD = 50

EXIT_CODE_PASSED = 0
EXIT_CODE_NOT_PASSED = 1
EXIT_CODE_ERROR = 2

ATTACK_STATE_QUEUED = 0
ATTACK_STATE_RUNNING = 1
ATTACK_STATE_COMPLETED = 2
ATTACK_STATE_ERRORED = -1

ENABLE_DEV_FEATURES = os.getenv("MINDGARD_DEV_FEATURES", "False").lower() == "true"



# Standard library imports
import logging
import sys
import time
from typing import Any, Callable

# Third party imports
from ratelimit import RateLimitException, limits

if sys.version_info >= (3, 10):
    # Standard library imports
    from typing import ParamSpec, TypeVar
else:
    # Third party imports
    from typing_extensions import ParamSpec, TypeVar

logger = logging.getLogger(__name__)

P = ParamSpec("P")
R = TypeVar("R")


def throttle(
    f: Callable[P, R],
    rate_limit: int,
    sleeper: Callable[[int], None] = time.sleep,
    clock: Callable[[], float] = time.monotonic,
) -> Callable[P, R]:
    """
    Throttle a function to a maximum number of calls per minute.

    Operates on a tumbling window of 60 seconds, during which if
    `rate_limit` is reached, the call is delayed until the window resets.

    Warning: this will release all throttled calls at once after the window resets.

    Args:
        f: the function (call-able) to throttle
        rate_limit: The number of calls allowed per minute
        sleeper: A sleep(n) compatible implementation [optional]
        clock: A time.monotonic() compatible implementation [optional]
    """
    ratelimited = limits(calls=rate_limit, clock=clock, period=60)

    def wrapper(*args: Any, **kwargs: Any) -> R:
        while True:
            try:
                return ratelimited(f)(*args, **kwargs)  # type: ignore # ratelimited is a wrapper around f
            except RateLimitException as exception:
                logger.debug("Rate limit reached, sleeping for %s seconds", exception.period_remaining)
                # loop once the window resets
                # add 2 seconds to the sleep to avoid resending before window has actually closed (rounding errors)
                sleeper(int(exception.period_remaining) + 2)  # type: ignore # there is no type on the received exception from lib

    return wrapper



# Standard library imports
import logging
import time
import uuid
from typing import Optional
from urllib.parse import urljoin

# Third party imports
import requests
from pydantic import BaseModel

# Project imports
from mindgard.version import VERSION

from .exceptions import ClientException

logger = logging.getLogger(__name__)


class PopRequest(BaseModel):
    """
    Request model for popping the next event in a user attack session.

    Attributes:
        runId (uuid.UUID): The unique identifier for the run/session.
    """

    runId: uuid.UUID


class PromptRequestEntry(BaseModel):
    """
    Represents a single prompt entry for OpenAI chat completion.

    Attributes:
        content (str): The text content of the prompt.
        role (str): The role (e.g., 'user', 'assistant').
    """

    content: str
    role: str


class OpenAIChatCompletion(BaseModel):
    """
    Container for OpenAI chat completion history.

    Attributes:
        openai (Optional[list[PromptRequestEntry]]): List of prompt entries for OpenAI chat.
    """

    openai: Optional[list[PromptRequestEntry]] = None


class Event(BaseModel):
    """
    Represents a prompt event in a user attack session.

    Attributes:
        id (uuid.UUID): Unique event identifier.
        sessionID (uuid.UUID): Session identifier.
        referrerID (uuid.UUID): Referrer identifier.
        attackName (str): Name of the attack.
        turnIndex (int): Index of the turn in the session.
        query (str): The prompt/query text.
        history (Optional[OpenAIChatCompletion]): Chat history for context.
    """

    id: uuid.UUID
    sessionID: uuid.UUID
    referrerID: uuid.UUID
    attackName: str
    turnIndex: int
    query: str
    history: Optional[OpenAIChatCompletion] = None


class PopResponse(BaseModel):
    """
    Response model for a pop event request.

    Attributes:
        runID (uuid.UUID): The run/session identifier.
        runStatus (str): Status of the run.
        event (Optional[Event]): The event data, if available.
    """

    runID: uuid.UUID
    runStatus: str
    event: Optional[Event] = None


class ResultPushRequest(BaseModel):
    """
    Request model for pushing the result of a prompt event.

    Attributes:
        requestEventID (uuid.UUID): Event ID being responded to.
        sessionID (uuid.UUID): Session identifier.
        referrerID (uuid.UUID): Referrer identifier.
        attackName (str): Name of the attack.
        turnIndex (int): Index of the turn.
        response (str): The model's response.
        durationMs (float): Duration in milliseconds.
        errorCode (Optional[int]): Error code, if any.
        errorMessage (Optional[str]): Error message, if any.
    """

    requestEventID: uuid.UUID
    sessionID: uuid.UUID
    referrerID: uuid.UUID
    attackName: str
    turnIndex: int
    response: str
    durationMs: float
    errorCode: Optional[int] = None
    errorMessage: Optional[str] = None


class AttackPromptEventsClient:
    """
    Client for interacting with the Prompt Events API.

    Provides methods to pop (retrieve) and push (submit) prompt events for user attack sessions.
    Handles authentication, request formatting, and error handling for communication with the backend service.

    Args:
        base_url (str): The base URL of the API endpoint.
        access_token (str): Bearer token for authentication.
    """

    def __init__(self, base_url: str, access_token: str):
        self._base_url = base_url
        self._pop_events_url = urljoin(self._base_url + "/", "user-attack-sessions/events/pop")
        self._push_events_url = urljoin(self._base_url + "/", "user-attack-sessions/events/push")
        self._access_token = access_token
        self._session = requests.Session()
        self._session.headers.update(
            {
                "Authorization": f"Bearer {self._access_token}",
                "Content-Type": "application/json",
                "User-Agent": f"mindgard-cli/{VERSION}",
                "X-User-Agent": f"mindgard-cli/{VERSION}",
            }
        )

    def pop(self, request: PopRequest) -> Optional[PopResponse]:
        """
        Retrieve the next available prompt event for a given run ID.

        Args:
            request (PopRequest): The request containing the run ID.

        Returns:
            Optional[PopResponse]: The response containing the event, or None if not found.

        Raises:
            ClientException: If the request fails or response validation fails.
        """
        response = self._session.post(self._pop_events_url, data=request.model_dump_json())
        if response.status_code == 404:
            logger.debug(f"No events found {request.model_dump()=} {response.text=} {response.status_code=}")
            return None

        if response.status_code != 200:
            logger.debug(f"Failed pop request {request.model_dump()=} {response.text=} {response.status_code=}")
            raise ClientException(message=response.text, status_code=response.status_code)

        try:
            return PopResponse.model_validate(response.json())
        except Exception as e:
            raise ClientException("problem validating PopResponse", response.status_code) from e

    def push(self, request: ResultPushRequest):
        """
        Submit the result of a prompt event, including response, timing, and error info.

        Args:
            request (ResultPushRequest): The request containing result data to push.

        Raises:
            ClientException: If the push request fails.
        """
        response = self._session.post(self._push_events_url, data=request.model_dump_json())

        if response.status_code != 201:
            logger.debug(f"Failed to push results {request.model_dump()=} {response.text=} {response.status_code=}")
            raise ClientException(message=str(response.json()), status_code=response.status_code)



# Standard library imports
import logging
import uuid
from typing import Any, Optional
from urllib.parse import urljoin

# Third party imports
from pydantic import BaseModel

# Project imports
from mindgard.api.recon.recon_prompt_events import ReconPromptEventsClient
from mindgard.api.recon.types import GetReconRequest, StartReconRequest, StartReconResponse

from ..exceptions import ClientException


class StartUnifiedReconRequest(BaseModel):
    type: str


class ReconResultMetadata(BaseModel):
    total: int
    total_detected: int
    results: dict[str, Any]


class ReconResult(BaseModel):
    id: uuid.UUID
    state: str
    result: Optional[ReconResultMetadata] = None
    reason: Optional[str] = None
    recommendation: Optional[str] = None
    project_id: str


class ReconClient(ReconPromptEventsClient):
    def __init__(self, base_url: str, access_token: str, project_id: str):
        super().__init__(base_url, access_token, project_id)
        self._base_url = base_url
        self._project_id = project_id

    def start_recon(self, request: StartUnifiedReconRequest) -> StartReconResponse:
        """
        Start a reconnaissance process with the Mindgard service

        Args:
            request (StartUnifiedReconRequest): Info (project ID) connecting this operation to a target name

        Raises:
            ClientException: Raised if the Mindgard service returns an error

        Returns:
            StartReconResponse: Mindgard service response from the start request
        """
        start_recon_url = urljoin(self._base_url + "/", f"projects/{self._project_id}/recon")
        response = self._session.post(start_recon_url, data=request.model_dump_json())

        if response.status_code != 201:
            logging.debug(f"Failed to start recon: {response.json()} - {response.status_code}")
            raise ClientException(response.json(), response.status_code)
        try:
            return StartReconResponse.model_validate(response.json())
        except Exception as e:
            raise ClientException("Problem parsing StartReconResponse", response.status_code) from e

    def get_recon_result(self, request: GetReconRequest) -> Optional[ReconResult]:
        """
        Get the reconnaissance result from the Mindgard service

        Args:
            request (GetReconRequest): Info connecting this operation to a reconnaissance session
        Raises:
            ClientException: Raised if the Mindgard service returns an error
        Returns:
            Optional[SystemPromptExtractResult]: Mindgard service response from the get request
        """
        get_recon_url = urljoin(self._base_url + "/", f"projects/{self._project_id}/recon/{request.recon_id}")

        response = self._session.get(get_recon_url)

        if response.status_code == 404:
            logging.debug(f"No reconnaissance found for recon_id={request.recon_id}: {response.text}")
            return None

        if response.status_code != 200:
            logging.debug(f"Failed to get recon result: {response.text} - {response.status_code}")
            raise ClientException(response.json(), response.status_code)

        try:
            return ReconResult.model_validate(response.json())
        except Exception as e:
            raise ClientException("Problem parsing ReconResult", response.status_code) from e



# Standard library imports
from typing import Any, Dict, Literal, Union

# Project imports
from mindgard.headers import parse_headers
from mindgard.wrappers.llm import LLMModelWrapper, get_llm_model_wrapper


# TODO: make an alias for model_type
def parse_args_into_model(args: Dict[str, Any]) -> LLMModelWrapper:
    return get_llm_model_wrapper(
        preset=args["preset"],
        headers=parse_headers(headers_comma_separated=args["headers"], headers_list=args["header"]),
        api_key=args["api_key"],
        url=args["url"],
        selector=args["selector"],
        request_template=args["request_template"],
        system_prompt=args["system_prompt"],
        model_name=args["model_name"],
        az_api_version=args["az_api_version"],
        tokenizer=args["tokenizer"],
        rate_limit=args["rate_limit"],
        force_multi_turn=args.get("force_multi_turn") or False,
    )



# Standard library imports
import dataclasses
from typing import Optional

# Third party imports
from jinja2 import Template


@dataclasses.dataclass
class Turn:
    prompt: str
    response: str


def render_messages_template(
    prompt: str, template: Template, system_prompt: Optional[str] = None, turns: list[Turn] = []
) -> str:
    return template.render(
        system_prompt=system_prompt,
        prompt=prompt,
        turns=[dataclasses.asdict(turn) for turn in turns],
    )



# Standard library imports
from concurrent.futures import ThreadPoolExecutor
from time import sleep
from typing import Any

# Third party imports
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn

# Project imports
from mindgard.api.exceptions import ClientException
from mindgard.api.recon.recon import ReconClient, ReconResult
from mindgard.auth import load_access_token
from mindgard.constants import API_BASE, DASHBOARD_URL
from mindgard.recon.constants import recon_constants
from mindgard.recon.recon_subcommand import ReconCommand
from mindgard.utils import CliResponse
from mindgard.wrappers.llm import LLMModelWrapper


def run_recon(type: str, console: Console, model_wrapper: LLMModelWrapper, final_args: dict[str, Any]):
    access_token = load_access_token()
    recon_service = ReconClient(API_BASE, access_token, final_args["project_id"])
    recon_command = ReconCommand(recon_service, model_wrapper, type)
    result: ReconResult = None

    def recon() -> None:
        nonlocal result
        recon_id = recon_command.start_recon(final_args["project_id"])
        recon_command.poll_recon(recon_id)
        result = recon_command.fetch_recon_result(recon_id)

    def recon_with_spinner() -> None:
        with Progress(
            SpinnerColumn(style="bold yellow"),
            TextColumn("[bold yellow]Probing target…"),
            transient=True,
            console=console,
        ) as progress:
            task_id = progress.add_task("probing", total=None)

            with ThreadPoolExecutor(max_workers=1) as executor:
                recon_future = executor.submit(recon)

                while not recon_future.done():
                    sleep(0.15)

                progress.update(task_id, completed=100)
                recon_future.result()

    # Helper to map recon type to constant name for dashboard link
    def reverse_lookup(type: str) -> str:
        for key, value in recon_constants._asdict().items():
            if value == type:
                return key
        return "unknown"

    try:
        recon_with_spinner()
    except ClientException as ex:
        raise ex

    if result is None:
        console.print(f"Failed to get result from {type} recon session")
        exit(CliResponse(1).code())

    console.print("\nProbing completed!", style="bold green")
    detected = result.result.total_detected > 0

    if not detected:
        console.print(f"\nNo clear signs of {type} from target system.", style="bold red")
        console.print(
            f"\n[underline]Reasoning[/underline]: The target didn't produce outputs that point to a successful {type}."
        )
        console.print(
            f"[underline]Recommendation[/underline]: Run the complete mindgard test suite for open-ended probing to uncover any additional risks. See [yellow]mindgard test --help[/yellow] to learn more.\n"
        )
        exit(CliResponse(0).code())

    console.print(
        f"\nFull results can be found here: {DASHBOARD_URL}/results/recon?project_id={result.project_id}&reconType={reverse_lookup(type)}\n"
    )



# Standard library imports
import logging
import uuid
from time import sleep
from typing import Optional

from ..api.recon.guardrail import (
    GetDetectResponse,
    GetFingerprintResponse,
    GetReconRequest,
    ReconGuardrailClient,
    StartDetectRequest,
    StartFingerprintRequest,
)
from ..api.recon.recon_prompt_events import (
    ALLOWED_EVENT_SUBJECTS,
    PopEventRequest,
    PopEventResponse,
    PromptResult,
    PushEventRequest,
)
from ..exceptions import MGException
from ..wrappers.llm import LLMModelWrapper


class GuardrailReconCommand:
    def __init__(self, service: ReconGuardrailClient, call_system_under_test: LLMModelWrapper):
        self.call_system_under_test = call_system_under_test
        self.service = service

    def start_detect(self, project_id: str) -> uuid.UUID:
        logging.debug(f"Starting guardrail detect reconnaissance for target: {project_id}")
        response = self.service.start_detect(StartDetectRequest(project_id=project_id))
        return response.recon_id

    def start_fingerprint(self, recon_id: uuid.UUID) -> uuid.UUID:
        f"Starting guardrail fingerprinting for target: {str(recon_id)}"
        response = self.service.start_fingerprint(StartFingerprintRequest(recon_id=recon_id))
        return response.recon_id

    def poll_detect(self, recon_id: uuid.UUID) -> None:
        return self.poll(recon_id, "GUARDRAIL_DETECT")

    def poll_fingerprint(self, recon_id: uuid.UUID) -> None:
        return self.poll(recon_id, "GUARDRAIL_FINGERPRINT")

    def poll(self, recon_id: uuid.UUID, event_subject: ALLOWED_EVENT_SUBJECTS) -> None:
        completed = False
        event: Optional[PopEventResponse] = None

        logging.debug(f"Polling for prompt requests to process: {str(recon_id)}")
        while not completed:
            completed, event = self._poll_inner(recon_id, event_subject)
            if completed:
                break  # dont sleep again if completed
            sleep(0.5)

        logging.debug(f"All prompt requests processed for recon_id: {str(recon_id)}")
        logging.debug("Fetching reconnaissance result...")

    def _poll_inner(self, recon_id: uuid.UUID, event_subject: ALLOWED_EVENT_SUBJECTS) -> tuple[bool, PopEventResponse]:
        request = PopEventRequest(
            source_id=recon_id,
            event_type=["prompt_request", "complete"],
            event_subject=event_subject,
        )

        event = self.service.pop(request)

        if event is None:
            return (False, None)
        if event.event_type == "complete":
            logging.debug(f"Reconnaissance completed for recon_id: {str(recon_id)}")
            return (True, event)

        if event.event_type == "prompt_request" and len(event.prompt_request) > 0:
            logging.debug(f"Processing prompt request")
            for request in event.prompt_request:
                response = None
                exception = None
                try:
                    response = self.call_system_under_test(request.prompt)
                except MGException as e:
                    logging.debug(f"Error communicating with target application: {e}")
                    exception = e

                content = response.response if response else None
                duration_ms = response.duration_ms if response else None
                error_code = None if exception is None else getattr(exception, "status_code", None)
                error_message = None if exception is None else str(exception)

                logging.debug(f"Pushing prompt result for request")
                push_event = PushEventRequest(
                    source_id=event.source_id,
                    event_type="prompt_result",
                    event_subject=event_subject,
                    prompt_result=[
                        PromptResult(
                            id=request.id,
                            content=content,
                            duration_ms=duration_ms,
                            prompt_request=request,
                            error_code=error_code,
                            error_message=error_message,
                        )
                    ],
                )

                self.service.push(push_event)
            return (False, event)

    def fetch_recon_detect_result(self, recon_id: uuid.UUID) -> GetDetectResponse:
        return self.service.get_detect_result(GetReconRequest(recon_id=recon_id))

    def fetch_recon_fingerprint_result(self, recon_id: uuid.UUID) -> Optional[list[GetFingerprintResponse]]:
        return self.service.get_fingerprint_result(GetReconRequest(recon_id=recon_id))



from .types import _SubparserType


def add_list_command(subparsers: _SubparserType):
    list_parser = subparsers.add_parser("list", help="List items")
    list_subparsers = list_parser.add_subparsers(dest="list_command")
    list_subparsers.add_parser("projects", help="List projects")
    return list_parser
